from __future__ import annotations
from typing import TYPE_CHECKING
from pandas import DataFrame
import os
import numpy as np
from typing import Optional, Dict, Tuple, List
from geopandas import GeoDataFrame, read_file
from geoprob_pipe.utils.statistics import calc_kar_waarde_lognormal, calc_kar_waarde_normal
import plotly.graph_objects as go
from geoprob_pipe.utils.validation_messages import BColors
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings
    from geoprob_pipe.cmd_app.parameter_input.input_parameter_tables import InputParameterTables


class InputParameterFigures:

    def __init__(self, app_settings: ApplicationSettings, tables: InputParameterTables, export: bool = False):
        self.app_settings: ApplicationSettings = app_settings
        self.tables: InputParameterTables = tables
        self.export: bool = export

        # Placeholders
        self.df_parameter_invoer: Optional[DataFrame] = None
        self.dict_vakindeling: Dict[int, Dict] = {}
        self.traject_length: Optional[float] = None
        self.x_min = 0
        self.x_max = 0  # Will be set
        self.y_min = 0
        self.y_max = 0  # Will be adjusted on each parameter

        # Perform logic
        self._gather_data()
        self._create_figures()

    def _gather_data(self):

        # Dataframe parameter invoer
        self.df_parameter_invoer = self.tables.df_parameter_invoer
        self.df_gis_join_parameter_invoer = self.tables.df_gis_join_parameter_invoer

        # Lengte traject
        gpkg_file_path = self.app_settings.geopackage_filepath
        gdf: GeoDataFrame = read_file(gpkg_file_path, layer="dijktraject")
        assert gdf.__len__() == 1
        self.traject_length = gdf.iloc[0].geometry.length
        self.x_max = self.traject_length

        # Vakindeling dictionary
        gdf_vakindeling: GeoDataFrame = read_file(self.app_settings.geopackage_filepath, layer="vakindeling")
        self.dict_vakindeling = gdf_vakindeling.set_index('id').to_dict(orient='index')

        # Uittredepunten dictionary
        gdf_uittredepunten: GeoDataFrame = read_file(self.app_settings.geopackage_filepath, layer="uittredepunten")
        self.dict_uittredepunten = gdf_uittredepunten.set_index('uittredepunt_id').to_dict(orient='index')

    @staticmethod
    def _get_display_values_from_row(row) -> Tuple[float, Optional[float], Optional[float]]:
        """

        :param row:
        :return: mean, 5% ondergrens, 95% bovengrens
        """

        # Deterministic
        mean_value: float = row['mean']
        if row['distribution_type'] == 'deterministic':
            return mean_value, None, None

        # Determine standard deviation
        variation = row['variation']
        deviation = row['deviation']
        deviation_value: Optional[float] = None
        if not np.isnan(deviation):
            deviation_value = deviation
        elif not np.isnan(variation):
            deviation_value = mean_value * variation
        else:
            ValueError(f"Should have either a variation or deviation. Or maybe a bug and contact the developer.")

        # Log normal
        if row['distribution_type'] == 'log_normal':
            kar_5pr = calc_kar_waarde_lognormal(mean=mean_value, sd=deviation_value, percentiel=0.05)  # TODO: Shift bepalen
            kar_95pr = calc_kar_waarde_lognormal(mean=mean_value, sd=deviation_value, percentiel=0.95)
            return mean_value, kar_5pr, kar_95pr

        # Normal
        if row['distribution_type'] == 'normal':
            kar_5pr = calc_kar_waarde_normal(mean=mean_value, std=deviation_value, percentiel=0.05)
            kar_95pr = calc_kar_waarde_normal(mean=mean_value, std=deviation_value, percentiel=0.95)
            return mean_value, kar_5pr, kar_95pr

        raise ValueError(f"Unknown distribution_type '{row['distribution_type']}'.")

    @staticmethod
    def _get_display_values_from_df(
            df: DataFrame
) -> Tuple[
        List[float],
        Optional[List[Optional[None]]],
        Optional[List[Optional[None]]],
    ]:
        """ Leest de gemiddelde en 5% en 95% karakteristieke waarden uit o.b.v. de mean en deviation. """

        mean_values = []
        kar5pr_values = []
        kar95pr_values = []

        for index, row in df.iterrows():
            mean_value = row['mean']
            mean_values.append(mean_value)
            if row['distribution_type'] == 'deterministic':
                kar5pr_values.append(None)
                kar95pr_values.append(None)
                continue
            deviation_value: Optional[float] = None
            if not np.isnan(row['deviation']):
                deviation_value = row['deviation']
            elif not np.isnan(row['variation']):
                deviation_value = mean_value * row['variation']
            else:
                ValueError(f"Should have either a variation or deviation. Or maybe a bug and contact the developer.")
            if row['distribution_type'] == 'log_normal':
                kar5pr_values.append(calc_kar_waarde_lognormal(
                    mean=mean_value, sd=deviation_value, percentiel=0.05))  # TODO: Shift bepalen
                kar95pr_values.append(
                    calc_kar_waarde_lognormal(mean=mean_value, sd=deviation_value, percentiel=0.95))
                continue
            if row['distribution_type'] == 'normal':
                kar5pr_values.append(calc_kar_waarde_normal(mean=mean_value, std=deviation_value, percentiel=0.05))
                kar95pr_values.append(calc_kar_waarde_normal(mean=mean_value, std=deviation_value, percentiel=0.95))
                continue

        return mean_values, kar5pr_values, kar95pr_values

    def _add_geospatial_level_data(self, fig: go.Figure, parameter_name: str) -> go.Figure:
        df_filter: DataFrame = self.df_gis_join_parameter_invoer[
            (self.df_gis_join_parameter_invoer['parameter'] == parameter_name) &
            (self.df_gis_join_parameter_invoer['scope'] == 'uittredepunt')]
        show_legend_item_mean = True

        for index, row in df_filter.iterrows():

            # Skip cdf curve
            if row['distribution_type'] == 'cdf_curve':
                continue

            # Add mean
            mean, kar_5pr, kar_95pr = self._get_display_values_from_row(row=row)
            x_value = self.dict_uittredepunten[row['scope_referentie']]['metrering']
            fig.add_trace(go.Scatter(
                x=[x_value],
                y=[mean],
                mode='markers',
                name="Geo-gerefereerd",
                legendgroup="Geo-gerefereerd",
                showlegend=show_legend_item_mean,
                marker=dict(color='rgba(0, 0, 117, 1)', size=10, symbol="circle")))
            show_legend_item_mean = False

            # Add (possibly) deviation
            if kar_5pr:
                fig.add_trace(go.Scatter(
                    x=[x_value, x_value],
                    y=[kar_5pr, kar_5pr],
                    mode='markers',
                    name="Geo-gerefereerd",
                    legendgroup="Geo-gerefereerd",
                    showlegend=False,
                    marker=dict(color='rgba(0, 0, 117, 1)', size=10, symbol="triangle-up")))
            if kar_95pr:
                fig.add_trace(go.Scatter(
                    x=[x_value, x_value],
                    y=[kar_95pr, kar_95pr],
                    mode='markers',
                    name="Geo-gerefereerd",
                    legendgroup="Geo-gerefereerd",
                    showlegend=False,
                    marker=dict(color='rgba(0, 0, 117, 1)', size=10, symbol="triangle-down")))

        return fig

    def _add_traject_level_data(self, fig: go.Figure, parameter_name: str) -> go.Figure:
        df_filter = self.df_parameter_invoer[
            (self.df_parameter_invoer['parameter'] == parameter_name) &
            (self.df_parameter_invoer['scope'] == 'traject')]
        for index, row in df_filter.iterrows():

            # Skip cdf curve
            if row['distribution_type'] == 'cdf_curve':
                continue

            mean, kar_5pr, kar_95pr = self._get_display_values_from_row(row=row)
            fig.add_trace(go.Scatter(
                x=[self.x_min, self.x_max, self.x_max, self.x_min, self.x_min],
                y=[self.y_min, self.y_min, mean, mean, self.y_min],
                fill='toself',
                mode='lines',
                fillcolor='lightblue',
                name="Traject-niveau",
                legendgroup="Traject-niveau",
                showlegend=True,
                line=dict(width=1, color="rgb(0, 0, 0)"),
            ))

            # Add (possibly) kar values
            if kar_5pr:
                fig.add_trace(go.Scatter(
                    x=[self.x_min, self.x_max],
                    y=[kar_5pr, kar_5pr],
                    mode='lines',
                    name="Traject-niveau",
                    legendgroup="Traject-niveau",
                    showlegend=False,
                    line=dict(width=4, color="rgba(0, 0, 0, 1)"),
                ))
            if kar_95pr:
                fig.add_trace(go.Scatter(
                    x=[self.x_min, self.x_max],
                    y=[kar_95pr, kar_95pr],
                    mode='lines',
                    name="Traject-niveau",
                    legendgroup="Traject-niveau",
                    showlegend=False,
                    line=dict(width=4, color="rgba(0, 0, 0, 1)", dash='dot'),
                ))

        return fig

    def _add_vak_level_data(self, fig: go.Figure, parameter_name: str) -> go.Figure:
        df_filter: DataFrame = self.df_parameter_invoer[
            (self.df_parameter_invoer['parameter'] == parameter_name) &
            (self.df_parameter_invoer['scope'] == 'vak') &
            (self.df_parameter_invoer['ondergrondscenario_naam'].isna())
        ]
        show_legend_item = True

        for index, row in df_filter.iterrows():

            # Skip cdf curve
            if row['distribution_type'] == 'cdf_curve':
                continue

            mean, kar_5pr, kar_95pr = self._get_display_values_from_row(row=row)
            vak_id = row['scope_referentie']
            x_min = self.dict_vakindeling[vak_id]['m_start']
            x_max = self.dict_vakindeling[vak_id]['m_end']

            # Add mean
            fig.add_trace(go.Scatter(
                x=[x_min, x_max, x_max, x_min, x_min],
                y=[self.y_min, self.y_min, mean, mean, self.y_min],
                fill='toself',
                mode='lines',
                fillcolor='orange',
                name="Vak-niveau",
                legendgroup="Vak-niveau",
                showlegend=show_legend_item,
                line=dict(width=1, color="rgba(0, 0, 0, 1)"),
            ))
            show_legend_item = False

            # Add (possibly) deviation
            if kar_5pr:
                fig.add_trace(go.Scatter(
                    x=[x_min, x_max],
                    y=[kar_5pr, kar_5pr],
                    mode='lines',
                    name="Vak-niveau",
                    legendgroup="Vak-niveau",
                    showlegend=False,
                    line=dict(width=4, color="rgba(0, 0, 0, 1)")))
            if kar_95pr:
                fig.add_trace(go.Scatter(
                    x=[x_min, x_max],
                    y=[kar_95pr, kar_95pr],
                    mode='lines',
                    name="Vak-niveau",
                    legendgroup="Vak-niveau",
                    showlegend=False,
                    line=dict(width=4, color="rgba(0, 0, 0, 1)", dash='dot')))

        return fig

    def _add_vak_level_per_scenario_data(self, fig: go.Figure, parameter_name: str) -> go.Figure:
        df_filter: DataFrame = self.df_parameter_invoer[
            (self.df_parameter_invoer['parameter'] == parameter_name) &
            (self.df_parameter_invoer['scope'] == 'vak') &
            (self.df_parameter_invoer['ondergrondscenario_naam'].notna())]
        show_legend_item = True

        for vak_id in df_filter['scope_referentie'].unique():
            df_filter2 = df_filter[df_filter['scope_referentie'] == vak_id]
            mean_values, kar_5pr_values, kar_95pr_values = self._get_display_values_from_df(df_filter2)
            max_mean = max(mean_values)
            x_min = self.dict_vakindeling[vak_id]['m_start']
            x_max = self.dict_vakindeling[vak_id]['m_end']

            # Add max mean with fill to self
            fig.add_trace(go.Scatter(
                x=[x_min, x_max, x_max, x_min, x_min],
                y=[self.y_min, self.y_min, max_mean, max_mean, self.y_min],
                fill='toself',
                mode='lines',
                fillcolor='rgb(148, 103, 189)',
                name="Vak-niveau per scenario",
                legendgroup="Vak-niveau per scenario",
                showlegend=show_legend_item,
                line=dict(width=1, color="rgba(0, 0, 0, 1)"),
            ))
            show_legend_item = False

            # Add (possibly) deviation
            for mean, kar_5pr_value, kar_95pr_value in zip(mean_values, kar_5pr_values, kar_95pr_values):
                if mean is not max_mean:
                    fig.add_trace(go.Scatter(
                        x=[x_min, x_max],
                        y=[mean, mean],
                        mode='lines',
                        name="Vak-niveau per scenario",
                        legendgroup="Vak-niveau per scenario",
                        showlegend=False,
                        line=dict(width=1, color="rgba(0, 0, 0, 1)")))
                if kar_5pr_value is not None:
                    fig.add_trace(go.Scatter(
                        x=[x_min, x_max],
                        y=[kar_5pr_value, kar_5pr_value],
                        mode='lines',
                        name="Vak-niveau per scenario",
                        legendgroup="Vak-niveau per scenario",
                        showlegend=False,
                        line=dict(width=4, color="rgba(0, 0, 0, 1)")))
                if kar_95pr_value is not None:
                    fig.add_trace(go.Scatter(
                        x=[x_min, x_max],
                        y=[kar_95pr_value, kar_95pr_value],
                        mode='lines',
                        name="Vak-niveau per scenario",
                        legendgroup="Vak-niveau per scenario",
                        showlegend=False,
                        line=dict(width=4, color="rgba(0, 0, 0, 1)", dash='dot')))

        return fig

    def _add_uittredepunt_level_data(self, fig: go.Figure, parameter_name: str) -> go.Figure:
        df_filter: DataFrame = self.df_parameter_invoer[
            (self.df_parameter_invoer['parameter'] == parameter_name) &
            (self.df_parameter_invoer['scope'] == 'uittredepunt')]
        show_legend_item_mean = True

        # print(f"{df_filter.columns=}")
        for index, row in df_filter.iterrows():

            # Skip cdf curve
            if row['distribution_type'] == 'cdf_curve':
                continue

            mean, kar_5pr, kar_95pr = self._get_display_values_from_row(row=row)
            x_value = self.dict_uittredepunten[row['scope_referentie']]['metrering']

            # Add mean
            fig.add_trace(go.Scatter(
                x=[x_value],
                y=[mean],
                mode='markers',
                name="Uittredepunt-niveau",
                legendgroup="Uittredepunt-niveau",
                showlegend=show_legend_item_mean,
                marker=dict(color='rgba(0, 0, 0, 1)', size=10, symbol="circle"),
            ))
            show_legend_item_mean = False

            # Add (possibly) deviation
            if kar_5pr:
                fig.add_trace(go.Scatter(
                    x=[x_value, x_value],
                    y=[kar_5pr, kar_5pr],
                    mode='markers',
                    name="Uittredepunt-niveau",
                    legendgroup="Uittredepunt-niveau",
                    showlegend=False,
                    marker=dict(color='rgba(0, 0, 0, 1)', size=10, symbol="triangle-up")))
            if kar_95pr:
                fig.add_trace(go.Scatter(
                    x=[x_value, x_value],
                    y=[kar_95pr, kar_95pr],
                    mode='markers',
                    name="Uittredepunt-niveau",
                    legendgroup="Uittredepunt-niveau",
                    showlegend=False,
                    marker=dict(color='rgba(0, 0, 0, 1)', size=10, symbol="triangle-down")))

        return fig

    @staticmethod
    def _add_legend_symbols(fig: go.Figure) -> go.Figure:
        fig.add_trace(go.Scatter(
            x=[-1000, -1000],
            y=[0, 0],
            mode='lines',
            name="95% bovengrens",
            legendgroup="95% bovengrens",
            showlegend=True,
            line=dict(width=4, color="rgba(0, 0, 0, 1)", dash='dot')))
        fig.add_trace(go.Scatter(
            x=[-1000, -1000],
            y=[0, 0],
            mode='lines',
            name="5% ondergrens",
            legendgroup="5% ondergrens",
            showlegend=True,
            line=dict(width=4, color="rgba(0, 0, 0, 1)")))
        fig.add_trace(go.Scatter(
            x=[-1000, -1000],
            y=[0, 0],
            mode='markers',
            name="95% bovengrens",
            legendgroup="95% bovengrens",
            showlegend=True,
            marker=dict(color='rgba(0, 0, 0, 1)', size=10, symbol="triangle-down")))
        fig.add_trace(go.Scatter(
            x=[-1000, -1000],
            y=[0, 0],
            mode='markers',
            name="5% ondergrens",
            legendgroup="5% ondergrens",
            showlegend=True,
            marker=dict(color='rgba(0, 0, 0, 1)', size=10, symbol="triangle-up")))

        return fig

    def _create_figures(self):
        export_dir = os.path.join(
            self.app_settings.workspace_dir,
            "exports",
            self.app_settings.datetime_stamp,
            "parameter_input_process")
        os.makedirs(export_dir, exist_ok=True)

        # TODO: Overal hovers toepassen

        parameters_to_iterate = self.df_gis_join_parameter_invoer['parameter'].unique().tolist()
        parameters_to_iterate.extend(self.df_parameter_invoer['parameter'].unique().tolist())
        for parameter_name in parameters_to_iterate:

            # Initiate figure
            fig = go.Figure()

            # Add geospatial-level
            self._add_geospatial_level_data(fig=fig, parameter_name=parameter_name)

            # Add traject-level
            fig = self._add_traject_level_data(fig=fig, parameter_name=parameter_name)

            # Add vak-level
            fig = self._add_vak_level_data(fig=fig, parameter_name=parameter_name)

            # Add vak-level per scenario
            fig = self._add_vak_level_per_scenario_data(fig=fig, parameter_name=parameter_name)

            # Add uittredepunten-level
            fig = self._add_uittredepunt_level_data(fig=fig, parameter_name=parameter_name)

            # Add some legend items
            fig = self._add_legend_symbols(fig=fig)

            # Update layout
            fig.update_layout(
                title=f"Parameter invoer voor '{parameter_name}'",
                xaxis_title="Metrering [m]",
                yaxis_title="Y-as")
            fig.update_layout(xaxis=dict(range=[0, self.x_max]))

            # Export figure
            if self.export:
                export_path = os.path.join(export_dir, f"parameter_invoer_{parameter_name}.html")
                fig.write_html(export_path, include_plotlyjs='cdn')

        if self.export:
            print(f"{BColors.UNDERLINE}Overzichtsfiguren van de invoertabellen zijn geëxporteerd naar:\n"
                  f"{export_dir}{BColors.ENDC}")



"""Deze module bevat verschillende geohydrologische methoden die worden
gebruikt bij de veiligheidsbeoordeling van dijken. Achtergrondinformatie
is te vinden in het 'Technisch Rapport Waterspanningen bij dijken'
:cite:t:`trw_2004`.
"""

import math


# Functions
def calc_lambda(kd: float, c: float) -> float:
    r""" Calculates leakage length

    .. math::
        \lambda = \sqrt{kDc}

    :param kd: transmissivity [m2/day]
    :param c: resistance of the topsoil [day]
    :return: leakage length lambda [m]
    """
    return math.sqrt(kd * c)


# W staat voor de effectieve voorland lengte formule, lam = lambda
# noinspection PyPep8Naming
def calc_W(lam: float, L: float) -> float:
    r"""Calculates effective leakage length by:

    .. math::

        W = \lambda tanh(\frac{L}{\lambda})

    Args:
        lam (float): leakage lengte [m]
        L (float): physical length [m]

    Returns:
        float: effective leakage length [m]
    """
    return lam * math.tanh(L / lam)


# noinspection PyPep8Naming
def calc_r_BIT(w1: float, l2: float, w3: float) -> float:
    r"""Calculates response in stationary models at inner toe based on
    given weights. See :cite:t:`trw_2004`.

    .. math::

        r_{BIT} = \frac{W_{3}}{W_{1} + L_{2} + W_{3}}

    Args:
        w1 (float): weight of foreland [m]
        l2 (float): length of dike base [m]
        w3 (float): weight of hinterland [m]

    Returns:
        float: response at inner toe [0.0-1.0]
    """
    return w3 / (w1 + l2 + w3)


# noinspection PyPep8Naming
def calc_r_BUT(w1: float, l2: float, w3: float) -> float:
    r"""Calculates response in stationary models at outer toe based on
    given weights.
    See :cite:t:`trw_2004`.

    .. math::

        r_{BUT} = \frac{L_{2} + W_{3}}{W_{1} + L_{2} + W_{3}}

    Args:
        w1 (float): weight of foreland [m]
        l2 (float): length of dike base [m]
        w3 (float): weight of hinterland [m]

    Returns:
        float: response at outer toe [0.0-1.0]
    """
    return (l2 + w3) / (w1 + l2 + w3)


def calc_respons2pot(
    h_ref: float, r_exit: float, h_riv: float
) -> float:  # Van respons naar potentiaal
    r"""Calculates potential from given response. See :cite:t:`trw_2004`.

    .. math::

        \phi(x) = h_{ref} + r(x) (h_{river} - h_{ref})

    Args:
        h_ref (float): potential in hinterland (polder)
        r_exit (float): response at given location [m]
        h_riv (float): water level at river

    Returns:
        float: potential at given location
    """
    return h_ref + r_exit * (h_riv - h_ref)


def calc_pot2response(
    phi: float, h_ref: float, h_riv: float
) -> float:  # Van potentiaal naar respons.
    r"""Calculates response from given potential. This is the reverse
    function of `calc_respons2pot`.

    .. math::

        r(x) = \frac{\phi(x)-h_{ref}}{h_{river} -h_{ref}}

    Args:
        phi (float): given potential [m+ref]
        h_ref (float): reference leve, e.g. potential in polder [m+ref]
        h_riv (float): water level at river, given potential [m+ref]

    Returns:
        float: response given reference level
    """
    return (phi - h_ref) / (h_riv - h_ref)


# noinspection PyPep8Naming
def calc_ang_frequency(T: float) -> float:
    r"""Calculates Angular frequency from period of a sinus wave.
    See :cite:t:`trw_2004` for background information.

    .. math::

        \omega = \frac{2 \pi}{T}

    Args:
        T (float): period [s]

    Returns:
        float: angular frequency [rad/s]
    """
    T_float = float(T)
    w = 2.0 * math.pi / T_float
    return w


# noinspection PyPep8Naming
def calc_P_from_T(T: float) -> float:
    r"""Calculates duration P from period T.

    .. math::

        P = \frac{T}{2}

    This is a helper function. Often a period T needs to be
    from a known storm surge P

    Args:
        T (float): Storm period T [s]

    Returns:
        float: Storm duration P [s]
    """
    return T / 2.0


# noinspection PyPep8Naming
def calc_T_from_P(P: float) -> float:
    r"""Calculates storm period from storm duration P.

    .. math::

        T = P * 2.0

    Args:
        P (float): Storm duration P [s]

    Returns:
        float: Storm period T [s]
    """
    return P * 2.0


# functies voor berekening cyclische spreidingslengte op basis van de
# stationaire spreidingslengte
# noinspection PyPep8Naming
def calc_lambda_cycl_from_stationary(
    LambdaStat: float, d: float, c_v: float, w: float
) -> float:
    r"""Calculates cyclic lambda from stationary leakage length

    see :cite:t:`bauduin_barends_1988`

    .. math::

        t_{h} = \frac{d^{2}}{c_{v}^{'}}

        \lambda_{\omega} = \frac{1.082 * \lambda_{s}}
        {\sqrt[4]{t_{h}^{'}\omega}}


    Args:
        LambdaStat (float): stationary leakage length [m]
        d (float): thickness of impermeable cover layer [m]
        c_v (float): one dimensional consolidation coëfficiënt  [m2/s]
        w (float): angular frequency [rad/s]

    Returns:
        float: cyclic lambda [m]
    """
    th = math.pow(d, 2.0) / c_v
    LambdaCycl = (1.082 * LambdaStat) / math.pow(th * w, 0.25)
    return LambdaCycl


# noinspection PyPep8Naming
def calc_lambda_cycl(LambdaCycl_1: float, T2: float, T1: float) -> float:
    r"""Calculates cyclic lambda from one period to another.
    See :cite:t:`trw_2004`.

    .. math::

        \lambda_{\omega, T_{2}}^{'} = \lambda_{\omega, T_{1}}^{'}
        \sqrt[4]{\frac{T_{2}}{T_{1}}}

    T1 and T2 are in same dimension (e.g. seconds, hours, days)

    Args:
        LambdaCycl_1 (float): known cyclic lambda at given period T1
        T2 (float): period for which cyclic lambda needs te be calculated
        T1 (float): known period

    Returns:
        float: _description_
    """
    return LambdaCycl_1 * math.pow(T2 / T1, 0.25)


def calc_theta(b: float, lambda_w_vl: float) -> float:
    r"""Calculates theta, see figure b4.13 from 'Technisch Rapport
    Waterspanningen bij dijken' :cite:t:`trw_2004`.
    Approximation by 5th degree polynomials.

    Args:
        b (float): with of river [m],
        lambda_w_vl (float): cyclic lambda of foreland [m]

    Raises:
        ValueError: Only for positive numbers

    Returns:
        float: theta from b4.13 :cite:t:`trw_2004`
    """
    x = b / lambda_w_vl
    if x < 0.0:
        raise ValueError
    elif x > 2.8:
        return 0.0
    else:
        return (
            0.0078 * math.pow(x, 5.0)
            - 0.082 * math.pow(x, 4.0)
            + 0.3139 * math.pow(x, 3.0)
            - 0.4683 * math.pow(x, 2.0)
            + 0.035 * x
            + 0.392699081698724
        )


def calc_f(b: float, lambda_w_vl: float) -> float:
    """Calculates f, see figure b4.13 from 'Technisch Rapport
    Waterspanningen bij dijken' :cite:t:`trw_2004`.
    Approximation by exponential function.

    Args:
        b (float): with of river [m],
        lambda_w_vl (float): cyclic lambda of foreland [m]

    Raises:
        ValueError: Only for positive numbers

    Returns:
        float: f from b4.13 :cite:t:`trw_2004`
        b           : with of river in m
        lambda_w_vl : cyclic lambda of foreland"""

    x = b / lambda_w_vl
    if x < 0.0:
        raise ValueError
    else:
        return 1.0 + 7.0659 * math.exp(-3.648 * x)


# noinspection PyPep8Naming
def calc_mean_pot_gradient(
    W1: float, W3: float, x_tp: float, mean_wl: float, phi_onv: float
) -> float:
    r"""Approximation of hydraulic head under daily (mean) conditions.
    Uses method from tipping point :cite:t:`trw_2004`.

    .. math::

       \phi(x) = h_{ref} + r(x) (h_{rivier} - h_{ref})

        r(x) = \frac{exp(\frac{- x_{tp}}{W_{3}})}{1+\frac{W_1}{W_3}}

    Args:
        W1 (float): weight of foreland under daily conditions [m]
        W3 (float): weight of hinterland under daily conditions [m]
        x_tp (float): given location from tipping point [m]
        mean_wl (float): daily water level at river
        phi_onv (float): potential in hinterland (polder)

    Returns:
        float: hydraulic head at given distance from tipping point
    """
    # Calculates respons tipping point
    rk = 1 / (1 + (W1 / W3))
    # Calculates response at given distance from tipping point
    r_uit = math.exp(-1 * (x_tp / W3))
    # Calculates hydraulic head at given distance from tipping point
    return phi_onv + (mean_wl - phi_onv) * rk * r_uit



from geoprob_pipe.calculations.limit_states.piping_lm import limit_state_wbi as system_variable_setup


# noinspection PyPep8Naming
def calc_Z_u(
        L_kwelweg: float,
        # Boundary condition parameters
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        top_zand: float, r_exit: float, k_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float, modelfactor_ff: float, modelfactor_3d: float,
        modelfactor_aniso: float, modelfactor_ml: float, i_c_h: float, r_c_deklaag: float,
        # Overige parameters
        gamma_water: float,
        # Constants  # TODO: Ombouwen tot globals
        d70_m: float, gamma_korrel: float, v: float, theta: float, eta: float, g: float
) -> float:
    r"""Grenstoestandfunctie voor opbarsten (uplift).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor opbarsten
    """
    return system_variable_setup(
        L_kwelweg=L_kwelweg, buitenwaterstand=buitenwaterstand, polderpeil=polderpeil, mv_exit=mv_exit,
        top_zand=top_zand, r_exit=r_exit, k_wvp=k_wvp, D_wvp=D_wvp, d70=d70, gamma_sat_deklaag=gamma_sat_deklaag,
        modelfactor_u=modelfactor_u, modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d, modelfactor_aniso=modelfactor_aniso,
        modelfactor_ml=modelfactor_ml, i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, gamma_water=gamma_water,
        d70_m=d70_m, gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
    )[0]


# noinspection PyPep8Naming
def calc_Z_h(
        L_kwelweg: float,
        # Boundary condition parameters
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        top_zand: float, r_exit: float, k_wvp: float, D_wvp: float, d70: float, gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float, modelfactor_ff: float, modelfactor_3d: float,
        modelfactor_aniso: float, modelfactor_ml: float, i_c_h: float, r_c_deklaag: float,
        # Overige parameters
        gamma_water: float,
        # Constants  # TODO: Ombouwen tot globals
        d70_m: float, gamma_korrel: float, v: float, theta: float, eta: float, g: float
) -> float:
    r""" Wrapper over de grenstoestandfunctie voor heave zodat deze bruikbaar is voor de Probabilistic Library.

    Returns:
        float: Z waarde van de grenstoestandfunctie voor heave
    """
    return system_variable_setup(
        L_kwelweg=L_kwelweg, buitenwaterstand=buitenwaterstand, polderpeil=polderpeil, mv_exit=mv_exit,
        top_zand=top_zand, r_exit=r_exit, k_wvp=k_wvp, D_wvp=D_wvp, d70=d70, gamma_sat_deklaag=gamma_sat_deklaag,
        modelfactor_u=modelfactor_u, modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d, modelfactor_aniso=modelfactor_aniso,
        modelfactor_ml=modelfactor_ml, i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, gamma_water=gamma_water,
        d70_m=d70_m, gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
    )[1]


# noinspection PyPep8Naming
def calc_Z_p(
        L_kwelweg: float,
        # Boundary condition parameters
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        top_zand: float, r_exit: float, k_wvp: float, D_wvp: float, d70: float, gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float, modelfactor_ff: float, modelfactor_3d: float,
        modelfactor_aniso: float, modelfactor_ml: float, i_c_h: float, r_c_deklaag: float,
        # Overige parameters
        gamma_water: float,
        # Constants  # TODO: Ombouwen tot globals
        d70_m: float, gamma_korrel: float, v: float, theta: float, eta: float, g: float
) -> float:
    r"""Grenstoestandfunctie voor het mechanisme terugschrijdende erosie (piping).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor piping
    """
    return system_variable_setup(
        L_kwelweg=L_kwelweg, buitenwaterstand=buitenwaterstand, polderpeil=polderpeil, mv_exit=mv_exit,
        top_zand=top_zand, r_exit=r_exit, k_wvp=k_wvp, D_wvp=D_wvp, d70=d70, gamma_sat_deklaag=gamma_sat_deklaag,
        modelfactor_u=modelfactor_u, modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d, modelfactor_aniso=modelfactor_aniso,
        modelfactor_ml=modelfactor_ml, i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, gamma_water=gamma_water,
        d70_m=d70_m, gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
    )[2]



# noinspection PyPep8Naming
def calc_Z_project(
        L_kwelweg: float,
        # Boundary condition parameters
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        top_zand: float, r_exit: float, k_wvp: float, D_wvp: float, d70: float, gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float, modelfactor_ff: float, modelfactor_3d: float,
        modelfactor_aniso: float, modelfactor_ml: float, i_c_h: float, r_c_deklaag: float,
        # Overige parameters
        gamma_water: float,
        # Constants  # TODO: Ombouwen tot globals
        d70_m: float, gamma_korrel: float, v: float, theta: float, eta: float, g: float
) -> float:
    r"""Grenstoestandfunctie voor het mechanisme piping. Deze grenstoestandfunctie bepaalt de Z waarde 
    op basis van de afzonderlijke Z waarden van de mechanismen uplift, heave en terugschrijdende erosie door
    max(Z_u, Z_h, Z_p).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor piping
    """
    return system_variable_setup(
        L_kwelweg=L_kwelweg, buitenwaterstand=buitenwaterstand, polderpeil=polderpeil, mv_exit=mv_exit,
        top_zand=top_zand, r_exit=r_exit, k_wvp=k_wvp, D_wvp=D_wvp, d70=d70, gamma_sat_deklaag=gamma_sat_deklaag,
        modelfactor_u=modelfactor_u, modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d, modelfactor_aniso=modelfactor_aniso,
        modelfactor_ml=modelfactor_ml, i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, gamma_water=gamma_water,
        d70_m=d70_m, gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
    )[3]


MODEL_NAMES = {
    calc_Z_u.__name__: "Uplift",
    calc_Z_h.__name__: "Heave",
    calc_Z_p.__name__: "Piping",
}



from pathlib import Path
from typing import Optional

import pandas as pd


class FileSystem:
    """FileSystem class which creates a convenient overview of a given folder and the files within it"""
    # TODO Later Should Klein: Is het echt nodig om een 'FileSystem'-object te maken? Deze functies bestaan al toch?

    def __init__(self, path_folder: str | Path, extension: Optional[str] = None) -> None:
        """Initialize FileSystem instance which creates a convenient overview of a given folder and the files within it

        Args:
            path_folder (str | Path): path of folder for which a FileSystem instance will be created
            extension (Optional[str], optional): extension of specific files to include in the FileSystem instance.
                Defaults to None, which returns all files path_folder regardless of extension.
        """
        self.folderpath = self.validate_path(path_folder)

        self.files = pd.DataFrame()
        self.files["filepath"] = self.find_files_in_dir(self.folderpath, extension)
        self.files["filename"] = [file.name for file in self.files["filepath"]]


    @staticmethod
    def validate_path(path_to_check: str | Path) -> Path:
        """Checks if the given path exists

        Args:
            path_to_check (str | Path): path to either a file or a folder

        Raises:
            FileNotFoundError: raised if the file/folder was not found

        Returns:
            Path: Path object for the validated path
        """
        if not Path(path_to_check).exists():
            raise FileNotFoundError(f"Path {path_to_check} doesn't exist")
        else:
            return Path(path_to_check).resolve()

    @staticmethod
    def find_files_in_dir(path_to_dir: str | Path, extension: str | None = None) -> list[Path]:
        """Find all files with a specified extension in a folder and its subfolders.
        Note: files starting with ~$ are ignored since these are generally temporary files

        Args:
            path_to_dir (str | Path): folder to search in
            extension (str, None): extension without the dot (e.g. "txt" instead of ".txt").
                                   If None is passed all files in dir, regardless of extension, are returned

        Returns:
            list[Path]: list of paths to the found files in the dir. Returns None if no files were found.
        """
        if isinstance(extension, str):
            files = [file
                     for file in Path(path_to_dir).resolve().rglob(f"*.{extension}")
                     if file.is_file() and not file.name.startswith("~$")]
        else:
            files = [file
                     for file in Path(path_to_dir).resolve().rglob(f"*")
                     if file.is_file() and not file.name.startswith("~$")]
        return files

    @staticmethod
    def find_subfolders(path_to_dir: str | Path) -> list[Path]:
        """Find all subfolders within a parent folder

        Args:
            path_to_dir (str | Path): folder to search in

        Returns:
            list[Path]: list of paths to the found subfolders in the dir.
        """
        return [subfolder for subfolder in Path(path_to_dir).iterdir() if subfolder.is_dir()]



from __future__ import annotations
from geopandas import GeoDataFrame, read_file
from pandas import DataFrame, read_sql_query
import sqlite3
from geoprob_pipe.cmd_app.spatial_joins.utils import append_to_gis_join_parameter_invoer_table
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def coupled_mv_exit_to_gis_parameter_invoer_table(app_settings: ApplicationSettings) -> bool:

    # Check if already added, if so skip
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    df_existing = read_sql_query("SELECT * FROM gis_join_parameter_invoer;", conn)
    if "mv_exit" in df_existing['parameter'].unique():
        return True

    # Getting necessary GeoDataframes
    gdf_exit_points: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")

    # Append to new version of geospatial storage
    df_to_append: DataFrame = gdf_exit_points[["mv_exit", "uittredepunt_id"]]
    df_to_append = df_to_append.rename(columns={"mv_exit": "mean"})
    append_to_gis_join_parameter_invoer_table(
        df_sjoin=df_to_append,
        parameter_name="mv_exit",
        geopackage_filepath=app_settings.geopackage_filepath)

    return True



from __future__ import annotations
# noinspection PyPep8Naming
from InquirerPy.prompts.input import InputPrompt as inq_text
import os
from geoprob_pipe.utils.validation_messages import BColors
from geoprob_pipe.cmd_app.comparisons import ComparisonCollector


def specify_dir_for_first_file():
    filepath: str = ""
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inq_text(
            message="Specificeer het volledige bestandspad naar het eerste .geoprob_pipe.gpkg-bestand.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not filepath.endswith(".geoprob_pipe.gpkg"):
            print(BColors.WARNING,
                  f"Het bestand moet een .geoprob_pipe.gpkg-bestand zijn. Jouw invoer "
                  f"{os.path.basename(filepath)} eindigt niet op deze extensie.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, "Het opgegeven bestandspad bestaat niet.",
                  BColors.ENDC)
            continue

        filepath_is_valid = True

    return filepath


def specify_dir_for_second_file():
    filepath: str = ""
    filepath_is_valid = False
    while filepath_is_valid is False:
        filepath: str = inq_text(
            message="Specificeer het volledige bestandspad naar het tweede .geoprob_pipe.gpkg-bestand.",
        ).execute()

        filepath = filepath.replace('"', '')

        if not filepath.endswith(".geoprob_pipe.gpkg"):
            print(BColors.WARNING,
                  f"Het bestand moet een .geoprob_pipe.gpkg-bestand zijn. Jouw invoer "
                  f"{os.path.basename(filepath)} eindigt niet op deze extensie.", BColors.ENDC)
            continue
        if not os.path.exists(filepath):
            print(BColors.WARNING, "Het opgegeven bestandspad bestaat niet.",
                  BColors.ENDC)
            continue

        filepath_is_valid = True

    return filepath


def specify_dir_for_comparison():
    workspace_dir: str = ""
    workspace_dir_is_valid = False
    while workspace_dir_is_valid is False:
        workspace_dir: str = inq_text(
            message="Specificeer het volledige pad naar de map waar je"
            + " de export van de vergelijking wilt opslaan.",
        ).execute()
        workspace_dir = workspace_dir.replace('"', '')

        if not os.path.exists(workspace_dir):
            print(BColors.WARNING, "De opgegeven map bestaat niet.",
                  BColors.ENDC)
            continue
        if not os.path.isdir(workspace_dir):
            print(BColors.WARNING, "De opgegeven locatie is geen map.",
                  BColors.ENDC)
            continue

        workspace_dir_is_valid = True

    return workspace_dir


def run_comparison(filepath1, filepath2, export_dir):
    print(BColors.OKBLUE, "Vergelijking wordt uitgevoerd.", BColors.ENDC)
    comparison = ComparisonCollector(filepath1, filepath2, export_dir)
    comparison.create_and_export_figures()
    print(BColors.OKBLUE,
          f"Vergelijking voltooid en opgeslagen in {comparison.export_dir}.",
          BColors.ENDC)


def start_comparison():
    filepath1 = specify_dir_for_first_file()
    filepath2 = specify_dir_for_second_file()
    export_dir = specify_dir_for_comparison()
    run_comparison(filepath1, filepath2, export_dir)



from geopandas import GeoDataFrame
from shapely import LineString, MultiLineString


def convert_mls_geom_column_to_ls(gdf: GeoDataFrame) -> GeoDataFrame:
    """ Converts all MultiLineString geometries with a single line that single LineString. To simplify it. """

    def unwrap_ls_in_mls(geom):
        """ Unwrap a single LineString in a MultiLineString, if it is indeed a single line. """

        if isinstance(geom, LineString):
            return geom

        if isinstance(geom, MultiLineString):
            mls: MultiLineString = geom
            if mls.geoms.__len__() > 1:
                return mls
            elif mls.geoms.__len__() == 0:
                return mls.geoms[0]

        raise NotImplementedError(
            f"This function was built to handle LineStrings or MultiLineStrings. Given geometry "
            f"is of type '{type(geom)}'. Please contact the developer.")

    gdf["geometry"] = gdf["geometry"].apply(unwrap_ls_in_mls)
    return gdf



from __future__ import annotations
import os
import plotly.graph_objects as go
import pandas as pd
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def phreatic_waterline(geoprob_pipe: GeoProbPipe, export: bool = False):
    # Prepare data
    df = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(
        filter_deterministic=False, filter_derived=False)
    df = df[["uittredepunt_id", "ondergrondscenario_id", "vak_id",
             "variable", "distribution_type", "physical_value"]]

    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf
    df = df.merge(
        gdf_uittredepunten[["uittredepunt_id", "metrering"]],
        on="uittredepunt_id", how="left"
    )

    # Determine scenario rank per uittredepunt_id
    # Assign "scenario_order" = 1, 2, 3, ... based on encounter order
    df["scenario_order"] = (
        df.groupby("uittredepunt_id")["ondergrondscenario_id"]
        .transform(lambda x: pd.factorize(x)[0] + 1)
    )

    scenario_orders = sorted(df["scenario_order"].unique())

    # Plotly setup
    fig = go.Figure()
    buttons = []

    colors = {
        "buitenwaterstand": "blue",
        "phi_exit": "green",
        "h_exit": "black",
        "top_zand": "brown"
    }
    symbols = {
        "buitenwaterstand": "square",
        "phi_exit": "x",
        "h_exit": "circle",
        "top_zand": "square"
    }
    names = {
        "buitenwaterstand": "Buitenwaterstand",
        "phi_exit": "Stijghoogte uittredepunt (phi_exit)",
        "h_exit": "Gehanteerde hoogte uittredepunt (h_exit)",
        "top_zand": "Bovenkant watervoerend pakket (top_zand)"
    }

    # Build one frame per scenario order
    for i, scen_order in enumerate(scenario_orders):
        df_case: pd.DataFrame = df[df["scenario_order"] == scen_order]

        # Add 4 variable traces per scenario_order
        for variable in ["buitenwaterstand", "phi_exit", "h_exit", "top_zand"]:
            df_var = df_case[df_case["variable"] == variable]
            dist_type = (str(df_var["distribution_type"].unique())
                         .strip("[]").strip("'"))
            fig.add_trace(go.Scatter(
                x=df_var['metrering'],
                y=df_var["physical_value"],
                mode='markers',
                name=f"{names[variable]} ({dist_type})",
                marker=dict(
                    symbol=symbols[variable],
                    size=5,
                    color=colors[variable]
                ),
                visible=(i == 0)
            ))

        # Button logic
        # Disappearance of title is a plotly bug
        total_traces = len(scenario_orders) * 4
        vis = [False] * total_traces
        vis[i*4:(i+1)*4] = [True]*4

        buttons.append(dict(
            label=f"Scenario {scen_order}",
            method="update",
            args=[
                {"visible": vis},
                {"title": f"Physical values t.o.v. NAP voor scenario {scen_order}"
                 + "<br><sup>Punten zijn per uittredepunt</sup>"}
            ]
        ))

    # Layout and controls
    fig.update_layout(
        title=f"Physical values t.o.v. NAP voor scenario {scenario_orders[0]}"
        + "<br><sup>Punten zijn per uittredepunt</sup>",
        updatemenus=[dict(
            active=0,
            buttons=buttons,
            direction="down",
            showactive=True,
            x=1.05, y=1.15
        )],
        xaxis=dict(
            title="Metrering",
            showgrid=True, gridwidth=0.5, gridcolor="gray",
        ),
        yaxis=dict(
            title="Hoogte [m+NAP]",
            showgrid=True, gridwidth=0.5, gridcolor="gray",
        ),
        legend=dict(
            orientation="h",
            yanchor="bottom",
            y=-0.25,
            xanchor="center",
            x=0.5
        ),
        height=600,
    )

    # Exports
    if export:
        export_dir = os.path.join(
            geoprob_pipe.visualizations.graphs.export_dir,
            "grafiek_physical_values"
        )
        os.makedirs(export_dir, exist_ok=True)

        # Export interactive HTML
        fig.write_html(
            os.path.join(export_dir, "phreatic_waterline.html"), include_plotlyjs='cdn')

        # Export one PNG per scenario_order
        for scen_order in scenario_orders:
            df_case = df[df["scenario_order"] == scen_order]
            fig_case = go.Figure()
            for variable in ["buitenwaterstand", "phi_exit",
                             "h_exit", "top_zand"]:
                df_var = df_case[df_case["variable"] == variable]
                fig_case.add_trace(go.Scatter(
                    x=df_var["metrering"],
                    y=df_var["physical_value"],
                    mode="markers",
                    name=names[variable],
                    marker=dict(
                        color=colors[variable],
                        symbol=symbols[variable],
                        size=5
                    )
                ))
            fig_case.update_layout(
                title=f"Phreatic waterline for Scenario {scen_order}",
                xaxis_title="Metrering",
                yaxis_title="Hoogte [m+NAP]",
                showlegend=True,
            )

    return fig



from geoprob_pipe.calculations.systems.wbi.limit_state_functions import (
    system_variable_setup, calc_Z_h, calc_Z_p, calc_Z_u, calc_Z_project)
from geoprob_pipe.calculations.systems.base_objects.system_calculation import SystemCalculation
from typing import List, Dict, Union, Tuple


class WBICalculation(SystemCalculation):
    """ Vooraf gedefinieerde System Reliability Calculation voor piping met het WBI-stijghoogtemodel. """

    def __init__(
            self,
            distributions: List[Dict],
            correlations: List[Tuple[str, str, float]] = None,
            reliability_settings: Dict[str, Union[str, float, int]] = None
    ):
        super().__init__(
            distributions=distributions, correlations=correlations, reliability_settings=reliability_settings)
        self.setup.variables_function = system_variable_setup
        self.setup.system_limit_states = [calc_Z_u, calc_Z_h, calc_Z_p]
        self.setup.project_limit_state = calc_Z_project



import os.path
import scipy.stats as sct
import sqlite3
from typing import Tuple, Optional
from shapely import LineString
import importlib.resources
from geopandas import read_file, GeoDataFrame


def _get_traject_id(hrd_path: str, hlcd_path: str) -> Tuple[int, str]:
    """ Queries first the HRD for the integer ID of the traject.
    Then queries the HLCD to find the textual traject ID.
    """
    conn = sqlite3.connect(hrd_path)
    cursor = conn.cursor()
    query = "SELECT TrackID FROM General;"
    cursor.execute(query)
    track_id = cursor.fetchone()[0]
    conn.close()

    conn = sqlite3.connect(hlcd_path)
    cursor = conn.cursor()
    query = f"SELECT Name FROM Tracks WHERE TrackID={track_id};"
    cursor.execute(query)
    traject_id = cursor.fetchone()[0]
    conn.close()

    return track_id, traject_id


def _query_dijktrajecten(traject_id: str):
    """ Queries the dijktrajecten.shp that was retrieved from the API of
    the Waterveiligheidsportaal [1].

    [1] https://www.nationaalgeoregister.nl/geonetwork/srv/dut/catalog.search#/metadata/fa4cc54e-26b3-4f25-b643-59458622901c
    """
    with importlib.resources.path(
            package='geoprob_pipe.input_data.dijktrajecten',
            resource='dijktrajecten.shp') as shp_path:
        gdf: GeoDataFrame = read_file(shp_path)
    gdf = gdf[gdf['TRAJECT_ID'] == traject_id]
    assert gdf.__len__() == 1, ("Only one traject id should have been found."
                                " Address this issue.")
    geom: LineString = gdf.iloc[0].geometry
    return gdf.iloc[0]['NORM_SW'], gdf.iloc[0]['NORM_OG'], geom.length


class TrajectNormering:
    """ Gathers the traject id and calculates the traject normering
    from the HRD-files.
    """

    def __init__(
            self,
            hrd_path: str,
            # Non-existent in HRD? For now as optional parameter.
            traject_naam: Optional[str] = None,
            norm_is_ondergrens: bool = True,
            # Where to find this? For now as manual parameter
            bovenrivierengebied: bool = True,
    ):

        # Input
        self.traject_naam: Optional[str] = traject_naam
        self.hrd_path = hrd_path
        self.hlcd_path = os.path.join(os.path.dirname(self.hrd_path),
                                      "hlcd.sqlite")
        self.bovenrivierengebied: bool = bovenrivierengebied

        # Parameters
        self.traject_id: str = _get_traject_id(self.hrd_path,
                                               self.hlcd_path)[1].strip()
        signaleringswaarde, ondergrens, traject_lengte = (
            _query_dijktrajecten(self.traject_id))
        self.signaleringswaarde: int = signaleringswaarde
        self.ondergrens: int = ondergrens
        self.w: float = 0.24
        self.traject_lengte: float = traject_lengte
        self.faalkanseis_signaleringswaarde = 1.0 / self.signaleringswaarde
        self.faalkanseis_ondergrens = 1.0 / self.ondergrens
        self.faalkanseis_norm = self.faalkanseis_ondergrens
        if not norm_is_ondergrens:
            self.faalkanseis_norm = self.signaleringswaarde
        self.beta_norm = sct.norm.ppf(self.faalkanseis_norm)
        self.n_dsn = 1 + (0.9 * self.traject_lengte) / 300.0
        if not self.bovenrivierengebied:
            self.n_dsn = 1 + (0.4 * self.traject_lengte) / 300.0
        # TODO Nu Must Klein: Eigenlijk hoofdletter N_dsn.
        # Maar ipv afkorting naam gebruiken?
        self.faalkanseis_sign_dsn = (
            self.w * self.faalkanseis_signaleringswaarde) / self.n_dsn
        self.beta_sign_dsn = sct.norm.ppf(self.faalkanseis_sign_dsn)
        self.faalkanseis_ond_dsn = (
            self.w * self.faalkanseis_ondergrens) / self.n_dsn
        self.beta_ond_dsn = sct.norm.ppf(self.faalkanseis_ond_dsn)
        self.beta_categorie_grenzen = {
            "I": [-1 * sct.norm.ppf(self.faalkanseis_sign_dsn / 30), 50],
            "II": [
                -1 * sct.norm.ppf(self.faalkanseis_sign_dsn),
                -1 * sct.norm.ppf(self.faalkanseis_sign_dsn / 30),
            ],
            "III": [
                -1 * sct.norm.ppf(self.faalkanseis_ond_dsn),
                -1 * sct.norm.ppf(self.faalkanseis_sign_dsn),
            ],
            "IV": [
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens),
                -1 * sct.norm.ppf(self.faalkanseis_ond_dsn),
            ],
            "V": [
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens * 30),
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens),
            ],
            "VI": [
                -50,
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens * 30),
            ],
        }
        self.riskeer_categorie_grenzen = {
            "+III": [-1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 1000),20],
            "+II": [
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 100),
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 1000)
            ],
            "+I": [
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 10),
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 100)
            ],
            "0": [
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde),
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde / 10)
            ],
            "-I": [
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens),
                -1 * sct.norm.ppf(self.faalkanseis_signaleringswaarde),
            ],
            "-II": [
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens * 10),
                -1 * sct.norm.ppf(self.faalkanseis_ondergrens)
            ],
            "-III": [2, -1 * sct.norm.ppf(self.faalkanseis_ondergrens * 10)],
        }



from probabilistic_library import (
    ReliabilityProject, DesignPoint, CombineProject, ReliabilityMethod,
    CombinerMethod, CombineType, Stochast, Settings)
from typing import Optional, Callable, List, Dict, Union, Tuple
from geoprob_pipe.utils.validation_messages import ValidationMessages
import logging


logger = logging.getLogger("geoprob-pipe")


DEFAULT_RELIABILITY_SETTINGS: Dict[str, Union[str, float, int]] = {
    "reliability_method": ReliabilityMethod.form.__str__(),
    "variation_coefficient": 0.02,
    "maximum_iterations": 1000,
    "relaxation_factor": 0.4,
}


class SystemSetup:

    def __init__(
            self,
            distributions: List[Dict],
            correlations: List[Tuple[str, str, float]],
            reliability_settings: Dict[str, Union[str, float, int]],

            # Following will be assigned in children
            variables_function: Optional[Callable] = None,
            system_limit_states: Optional[List[Callable]] = None,
            project_limit_state: Optional[Callable] = None,

    ):
        self.distributions: List[Dict] = distributions
        self.correlations: List[Tuple[str, str, float]] = correlations
        self.reliability_settings: Dict[str, Union[str, float, int]] = reliability_settings
        if self.reliability_settings is None or self.reliability_settings.__len__() == 0:
            self.reliability_settings = DEFAULT_RELIABILITY_SETTINGS
        if self.correlations is None:
            self.correlations = []

        # Assigned by children:
        self.variables_function: Optional[Callable] = variables_function
        self.system_limit_states: Optional[List[Callable]] = system_limit_states
        self.project_limit_state: Optional[Callable] = project_limit_state
        # TODO: At some point combine the variables function and the project_limit_state? I think that can be done.


class SystemResults:

    def __init__(self):

        self.reliability_project = ReliabilityProject()  # Initiated for later use
        # -> Used for all reliability project calculations, i.e. the system limit states and the project limit state.
        self.dps_limit_states: List[DesignPoint] = []  # Placeholder
        self.dp_reliability: Optional[DesignPoint] = None  # Placeholder

        self.combine_project = CombineProject()  # Initiated for later use
        self.dp_combine: Optional[DesignPoint] = None  # Placeholder


class SystemCalculation:
    """ Performs """

    def __init__(
            self,
            distributions: List[Dict],
            correlations: List[Tuple[str, str, float]],
            reliability_settings: Dict[str, Union[str, float, int]] = None,

            # Following will be assigned in children
            variables_function: Optional[Callable] = None,
            system_limit_states: Optional[List[Callable]] = None,
            project_limit_state: Optional[Callable] = None,
    ):

        if reliability_settings is None:
            reliability_settings = DEFAULT_RELIABILITY_SETTINGS


        self.validation_messages = ValidationMessages()
        self.metadata = {}

        self.setup = SystemSetup(
            distributions=distributions, correlations=correlations, reliability_settings=reliability_settings,
            variables_function=variables_function, system_limit_states=system_limit_states,
            project_limit_state=project_limit_state)
        self.results = SystemResults()  # Initiation of empty object

    def run(self):
        _apply_settings(self)
        _apply_distributions(self)
        _apply_correlations(self)
        _generate_dps_limit_states(self)
        _generate_dp_reliability(self)
        _generate_dp_combine(self)


def _apply_settings(system_calculation: SystemCalculation):
    """ Applies the settings of both the Reliability Project and the Combine Project, where the former can be user
    defined settings. """

    # Reliability Project
    for attr_name, value in system_calculation.setup.reliability_settings.items():
        if attr_name not in Settings().__dir__():
            raise ValueError(
                f"Attribute '{attr_name}' not found in the ReliabilityCalculation.Settings-class. "
                f"Available attributes are:\n"
                f"{Settings().__dir__()}")
        setattr(system_calculation.results.reliability_project.settings, str(attr_name), value)

    # Combine Project
    system_calculation.results.combine_project.settings.combiner_method = CombinerMethod.importance_sampling
    system_calculation.results.combine_project.settings.combine_type = CombineType.parallel


def _apply_distributions(system_calculation: SystemCalculation):
    """ Both initiates the variables and applies the distributions. """

    def _initiatie_variables():
        system_calculation.results.reliability_project.model = system_calculation.setup.variables_function

    def _system_variable_keys() -> List[str]:
        return_array = []
        for item in system_calculation.setup.distributions:
            return_array.append(item['name'])
        return return_array

    def _validate_distributions_are_provided():
        """ Validate all system variables have a distribution provided. """
        system_variable_keys = _system_variable_keys()
        for var_item in system_calculation.results.reliability_project.variables:
            var_item: Stochast
            if var_item.name not in system_variable_keys:
                raise KeyError(
                    f"The system variable '{var_item.name}' has no distribution provided in "
                    f"system_variable_distributions-list. Please do so before running the system.")

    def _apply_distributions_on_variables():
        for item in system_calculation.setup.distributions:
            name = item['name']

            # Check if variable exists
            if system_calculation.results.reliability_project.variables[name] is None:
                system_calculation.validation_messages.add_warning(
                    msg=f"The variable '{name}' is unknown in the ReliabilityProject, i.e. in the given "
                        f"'system_variables_setup'-function. For now this application skips unnecessary variables. If "
                        f"the variable is necessary, revisit your 'system_variables_setup'-function and the limit state "
                        f"functions.")
                # TODO Nu Should Klein: Feedback aan gebruiker dat er validation messages zijn.
                continue

            system_calculation.results.reliability_project.variables[name].distribution = item['distribution_type']

            # Key-worded arguments for uniform
            if 'minimum' in item.keys():
                system_calculation.results.reliability_project.variables[name].minimum = item['minimum']
            if 'maximum' in item.keys():
                system_calculation.results.reliability_project.variables[name].maximum = item['maximum']

            # Key-worded arguments for deterministic, normal and/or log_normal
            if 'mean' in item.keys():
                system_calculation.results.reliability_project.variables[name].mean = item['mean']
            if 'deviation' in item.keys():
                system_calculation.results.reliability_project.variables[name].deviation = item['deviation']
            if 'variation' in item.keys():
                system_calculation.results.reliability_project.variables[name].variation = item['variation']

            # Key-worded arguments for cdf-curve
            if 'fragility_values' in item.keys():
                system_calculation.results.reliability_project.variables[name].fragility_values.extend(
                    item['fragility_values'])

    _initiatie_variables()
    _validate_distributions_are_provided()
    _apply_distributions_on_variables()


def _apply_correlations(system_calculation: SystemCalculation):
    """ Correlations will be applied to the Reliability Project. """

    for correlation in system_calculation.setup.correlations:
        system_calculation.results.reliability_project.correlation_matrix[
            correlation[0],  # Parameter name A
            correlation[1]  # Parameter name B
        ] = correlation[2]  # Correlation value between 0.0 and 1.0


def _generate_dps_limit_states(system_calculation: SystemCalculation):
    """ Generates the design points for the limit states separately, i.e. uplift, heave and Sellmeijer. It uses the
    Reliability Project for this. """

    for model_callable in system_calculation.setup.system_limit_states:
        system_calculation.results.reliability_project.model = model_callable
        _apply_correlations(system_calculation=system_calculation)
        # noinspection PyBroadException
        try:
            system_calculation.results.reliability_project.run()
        except Exception:
            logger.error("Run failed in _generate_dps_limit_states")
        design_point = system_calculation.results.reliability_project.design_point
        if design_point is None:
            logger.debug(f"For limit state: {model_callable}")
            raise TypeError("Design_point is NoneType, run has failed.")
        design_point.identifier = model_callable.__name__
        system_calculation.results.dps_limit_states.append(design_point)


def _generate_dp_reliability(system_calculation: SystemCalculation):
    """ Generates the design point for the piping limit state, i.e. an encapsulating limit state for uplift, heave
    and Sellmeijer. It uses the Reliability Project for this. """
    model_callable: Callable = system_calculation.setup.project_limit_state
    system_calculation.results.reliability_project.model = model_callable
    _apply_correlations(system_calculation=system_calculation)
    # noinspection PyBroadException
    try:
        system_calculation.results.reliability_project.run()
    except Exception:
        logger.error("Run failed in _generate_dp_reliability")
    design_point = system_calculation.results.reliability_project.design_point
    if design_point is None:
        logger.debug(f"For limit state: {model_callable}")
        raise TypeError("Design_point is NoneType, run has failed.")
    design_point.identifier = model_callable.__name__
    system_calculation.results.dp_reliability = design_point


def _generate_dp_combine(system_calculation: SystemCalculation):
    """ Generates the design point for the Combine Project. """

    for design_point in system_calculation.results.dps_limit_states:
        system_calculation.results.combine_project.design_points.append(design_point)
    system_calculation.results.combine_project.run()
    system_calculation.results.combine_project.design_point.identifier = "combine"
    system_calculation.results.dp_combine = system_calculation.results.combine_project.design_point



from geoprob_pipe.cmd_app.spatial_joins.coupled_hrd import coupled_hrd_to_uittredepunten
from geoprob_pipe.cmd_app.spatial_joins.coupled_distance_to_exit_points import (
    coupled_distances_to_uittredepunten)
from geoprob_pipe.cmd_app.spatial_joins.coupled_polderpeil import coupled_polderpeil_to_uittredepunten
from geoprob_pipe.cmd_app.spatial_joins.coupled_exit_points_to_refline import coupled_uittredepunten_to_refline
from geoprob_pipe.cmd_app.spatial_joins.coupled_exit_points_to_vakken import coupled_uittredepunten_to_vakken
from geoprob_pipe.cmd_app.spatial_joins.coupled_mv_exit import coupled_mv_exit_to_gis_parameter_invoer_table



from __future__ import annotations
from typing import TYPE_CHECKING
from pandas import DataFrame
from geoprob_pipe.visualizations.other.overview.generate_flow_chart_v2 import generate_overview_flow_chart_with_betas
import os
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Other:

    def __init__(self, app_obj: GeoProbPipe):
        self.geoprob_pipe = app_obj

    @property
    def export_dir(self) -> str:
        path = os.path.join(self.geoprob_pipe.visualizations.export_dir, "visualizations")
        os.makedirs(path, exist_ok=True)
        return path

    def export_visualizations(self):
        df = self.geoprob_pipe.results.df_beta_scenarios_final
        lowest_beta_row: DataFrame  = df.loc[df['beta'].idxmin()]
        generate_overview_flow_chart_with_betas(
            app_obj=self.geoprob_pipe,
            export_dir=self.export_dir,
            ondergrondscenario_id=lowest_beta_row['ondergrondscenario_id'],
            uittredepunt_id=lowest_beta_row['uittredepunt_id']
        )




CHANGELOG = {
    "1.5.3":
        "Implementatie van DEBUG-modus, Pythonic-logger en opslaan van log berichten in GeoPackage-bestand.",
    "1.4.1":
        "Documentation update.",
    "1.4.0":
        "Implementatie van een nieuwe versie van de System Calculation. Van af nu worden er drie scenario analyses "
        "uitgevoerd: een Combine Project en Reliability project over een gezamenlijke grenstoestandsfuncties voor "
        "uplift, heave en piping, en een Reliability Project over de grenstoestandsfuncties apart. Vervolgens wordt "
        "op basis van een stroomschema (zie documentatie) de meest optimale berekening geselecteerd.",
    "1.3.21":
        "Start met validatie van invoer Excel-sheets. Bewuste keuze voor SQL-queries (leesbaarheid), in plaats van "
        "pandera.",
    "1.3.20":
        "New feature: Methode geïmplementeerd om input_parameters.xlsx te valideren op invoer. Methode is voor nu "
        "een start. Validatie moet nog toegevoegd worden.",
    "1.3.19":
        "(a) Hotfix: Nieuwe versie probabilistic_library stond geen Beta van de overschrijdingsfrequentielijn toe "
        "van hoger dan 8.0. Hydra-NL lijkt tot en met 8.0 te gaan, waardoor het laatste punt soms net boven de 8.0 "
        "valt. Er wordt nu afgekapt in GeoProb-Pipe. "
        "(b) New feature: Indien een berekening faalt, wordt er nu een log opgeslagen van deze berekening. Applicatie "
        "blijft ook doordraaien, i.p.v. crash."
        "(c) Hotfix: Validatie issue bij importeren nieuwe Input-Excel opgelost.",
    "1.3.1":
        "Hotfix. Visualisatie en export issues. Geen issue aan berekeningen zelf. 1e issue: Foute beta waarden in de "
        "hover van de betrouwbaarheidsindex.html. 2e issue: De df_limit_states werd geëxporteerd als df_scenarios.",
    "1.3.0":
        "Probabilistic Library nu toegevoegd als 'setup dependency' nadat Deltares deze heeft vrijgegeven op PyPI. "
        "Eveneens Chrome uitgefaseerd als software requirement. Daardoor zijn beiden geen handmatige software "
        "requirements meer.",
    "1.2.0":
        "Implementatie van parallel rekenen en het inspecteren van een enkele berekening.",
    "1.1.0":
        "Implementatie van een eerste versie van het vergelijken van twee verschillende GeoProb-Pipe bestanden.",
    "1.0.0":
        "Implementatie GeoPackage als data (in- en uitvoer) bestand én implementatie van keuze in geohydrologische "
        "modellen.",
    "0.1.0":
        "Toevoegen features na interne gebruikerssessie WSRL 2025-08-11: (a) geopackage-export met resultaten, "
        "(b) Excel-export met physical values, alphas en invloedsfactoren, (c) Excel-export met resultaat per vak, "
        "(d) Excel-export met validatie berichten en (e) physical values in grafiek waterstandssfrequentielijn.",
    "0.0.15":
        "Eerste vrijgave van wheel installatie-bestand.",
}



from __future__ import annotations
from typing import TYPE_CHECKING
from geoprob_pipe.spatial.gdf_beta_limit_states import (
    get_gdf_beta_limit_states, get_gdf_beta_scenarios_final, get_gdf_beta_scenarios_cp, get_gdf_beta_scenarios_rp,
    get_gdf_beta_uittredepunten)
from geopandas import GeoDataFrame
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class Spatial:

    def __init__(self, geoprob_pipe: GeoProbPipe):
        self.geoprob_pipe = geoprob_pipe

    def get_gdf_beta_limit_states(self, export: bool = False) -> GeoDataFrame:
        return get_gdf_beta_limit_states(self.geoprob_pipe, export=export)

    def get_gdf_beta_scenarios_rp(self, export: bool = False) -> GeoDataFrame:
        return get_gdf_beta_scenarios_rp(self.geoprob_pipe, export=export)

    def get_gdf_beta_scenarios_cp(self, export: bool = False) -> GeoDataFrame:
        return get_gdf_beta_scenarios_cp(self.geoprob_pipe, export=export)

    def get_gdf_beta_scenarios_final(self, export: bool = False) -> GeoDataFrame:
        return get_gdf_beta_scenarios_final(self.geoprob_pipe, export=export)

    def get_gdf_beta_uittredepunten(self, export: bool = False) -> GeoDataFrame:
        return get_gdf_beta_uittredepunten(self.geoprob_pipe, export=export)

    @property
    def export_dir(self) -> str:
        return self.geoprob_pipe.input_data.app_settings.workspace_dir

    def export_geopackage(self):
        self.get_gdf_beta_limit_states(export=True)
        self.get_gdf_beta_scenarios_rp(export=True)
        self.get_gdf_beta_scenarios_cp(export=True)
        self.get_gdf_beta_scenarios_final(export=True)
        self.get_gdf_beta_uittredepunten(export=True)



from __future__ import annotations
from typing import TYPE_CHECKING
from geoprob_pipe.utils.validation_messages import BColors
from geopandas import GeoDataFrame, read_file
from pathlib import Path
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def coupled_uittredepunten_to_vakken(app_settings: ApplicationSettings) -> bool:

    # Read uittredepunten
    gdf_exit_points: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="uittredepunten")

    # Check if already added
    if "vak_id" in gdf_exit_points.columns:
        print(BColors.OKBLUE, f"✔  Vakken al gekoppeld aan uittredepunten.", BColors.ENDC)
        return True  # Assuming already added

    # Spatial analyses
    gdf_vakindeling: GeoDataFrame = read_file(app_settings.geopackage_filepath, layer="vakindeling")
    gdf_exit_points_with_vakken = gdf_exit_points.sjoin_nearest(
        gdf_vakindeling[['geometry', 'id']], how='left', distance_col='distance')

    # Define which columns to keep (after spatial join)
    columns_to_keep = list(gdf_exit_points.columns)
    columns_to_keep.append("id")
    gdf_new_exit_points = gdf_exit_points_with_vakken[columns_to_keep]
    gdf_new_exit_points = gdf_new_exit_points.rename(columns={"id": "vak_id"})

    # Store back in geopackage
    gdf_new_exit_points.to_file(Path(app_settings.geopackage_filepath), layer="uittredepunten", driver="GPKG")
    print(BColors.OKBLUE, f"✅  Vakken zijn nu gekoppeld aan de uittredepunten.", BColors.ENDC)
    return True



from __future__ import annotations
from pandas import DataFrame, concat
from probabilistic_library import DesignPoint, Alpha
from typing import TYPE_CHECKING, Dict, List, Union, cast, Optional
import numpy as np
from geoprob_pipe.calculations.systems.mappers.calculation_mapper import CALCULATION_MAPPER
if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe
    from geoprob_pipe.calculations.systems.base_objects.system_calculation import (
        SystemCalculation)
    from geoprob_pipe.calculations.systems.build_and_run import CalcResult


def collect_stochast_values(calc: SystemCalculation, df_scenario_final: DataFrame) -> DataFrame:
    """ Collects all Alphas, Influence factors and Physical values of the stochast input parameters. """

    # Get method used
    df = df_scenario_final.copy(deep=True)
    assert df.__len__() == 1, \
        f"Developer note: Assumption that this dataframe must have 1 row at all times. If triggered, then investigate."
    method_used = df.iloc[0]['method_used']

    # Create
    def create_df_rows_for_design_point(dp: DesignPoint, append: Optional[str] = None) -> List[Dict[str, Union[str, float]]]:
        rows_from_dp = []
        design_point_method = method_used
        if append:
            design_point_method = f"{design_point_method} ({append})"
        for alpha in dp.alphas:
            alpha: Alpha
            rows_from_dp.append({
                "uittredepunt_id": calc.metadata['uittredepunt_id'],
                "ondergrondscenario_id": calc.metadata['ondergrondscenario_naam'],
                "vak_id": calc.metadata['vak_id'],
                "design_point": design_point_method,
                "variable": alpha.identifier,
                "distribution_type": alpha.variable.distribution.value,
                "alpha": alpha.alpha,
                "influence_factor": alpha.alpha * alpha.alpha,
                "physical_value": alpha.x})
        return rows_from_dp

    rows = []

    if method_used == "3: Max Limit States":
        # Gather for separate limit states
        for design_point in calc.results.dps_limit_states:
            rows.extend(create_df_rows_for_design_point(dp=design_point, append=design_point.identifier))

    elif method_used == "2: Reliability Project":
        # Gather for reliability project
        dp_reliability = cast(DesignPoint, calc.results.dp_reliability)
        rows.extend(create_df_rows_for_design_point(dp=dp_reliability))

    elif method_used == "1: Combine Project":
        # Gather for combine project
        dp_combine = cast(DesignPoint, calc.results.dp_combine)
        rows.extend(create_df_rows_for_design_point(dp=dp_combine))

    else:
        raise ValueError(f"Unknown method_used '{method_used}'. Contact a developer.")

    # Generate df from rows
    df = DataFrame(rows)

    return df


def _combine_stochast_values(calc_results: List[CalcResult])  -> DataFrame:
    df = concat((result.df_stochast for result in calc_results), ignore_index=True)
    return df


def calculate_derived_values(df_scenarios_final: DataFrame, geohydrologisch_model: str):
    """ Re-calculates all derived physical values, i.e. intermediate values that were calculated inside the limit state
    functions. These are not returned by the probabilistic library, hence we need to re-calculate them. """

    # Get kwargs per calculation
    df = df_scenarios_final.copy(deep=True)
    assert df.__len__() == 1, \
        f"Developer note: Assumption that this dataframe must have 1 row at all times. If triggered, then investigate."
    method_used = df.iloc[0]['method_used']

    if method_used == "3: Max Limit States":
        df['physical_values'] = df['system_calculation'].apply(
            lambda sc: {alpha.variable.name: alpha.x
                        for dp in sc.results.dps_limit_states
                        for alpha in dp.alphas})

    elif method_used == "2: Reliability Project":
        df['physical_values'] = df['system_calculation'].apply(
            lambda sc: {alpha.variable.name: alpha.x for alpha in sc.results.dp_reliability.alphas})

    elif method_used == "1: Combine Project":
        df['physical_values'] = df['system_calculation'].apply(
            lambda sc: {alpha.variable.name: alpha.x for alpha in sc.results.dp_combine.alphas})
    else:
        raise NotImplementedError

    # Calculate the derived values
    def derived_values_single_calculation(model_naam: str, **kwargs):
        return_keys: List[str] = CALCULATION_MAPPER[model_naam]["system_return_parameter_keys"]
        system_limit_state_function = CALCULATION_MAPPER[model_naam]["limit_state_function"]
        derived_values = {key: value for key, value in zip(return_keys, system_limit_state_function(**kwargs))}
        return {**derived_values}

    df['derived_physical_values'] = df['physical_values'].apply(
        lambda kwargs: derived_values_single_calculation(model_naam=geohydrologisch_model, **kwargs))

    # Create df with row per derived physical value
    df_new = df[["uittredepunt_id", "ondergrondscenario_id", "vak_id", "derived_physical_values"]].copy(deep=True)
    df_new['design_point'] = method_used
    df_new['distribution_type'] = "derived"
    df_new['alpha'] = np.nan
    df_new['influence_factor'] = np.nan

    # Expand dictionary to new rows
    df_new = concat([
        DataFrame({
            **row.drop('derived_physical_values'),
            'variable': list(row['derived_physical_values'].keys()),
            'physical_value': list(row['derived_physical_values'].values())
        })
        for _, row in df_new.iterrows()
    ], ignore_index=True)

    return df_new


def _combine_derived_values(calc_results: List[CalcResult]) -> DataFrame:
    return concat((result.df_derived for result in calc_results), ignore_index=True)


def construct_df(geoprob_pipe: GeoProbPipe):

    # Merge derived and stochast values
    df = concat([
        _combine_stochast_values(geoprob_pipe.calc_results),
        _combine_derived_values(geoprob_pipe.calc_results)
    ])

    # Sort
    df = df.sort_values(by=["vak_id", "uittredepunt_id", "ondergrondscenario_id", "design_point", "variable"])
    return df.reset_index(drop=True)
    # TODO Later Could Klein: Bespreken of we de physical values willen afronden? Af wellicht afrondden in de export.



from __future__ import annotations
from typing import TYPE_CHECKING, List
from geoprob_pipe.calculations.systems.mappers.calculation_mapper import (
    CALCULATION_MAPPER)
from multiprocessing import Pool, cpu_count
import logging
from io import StringIO
import sqlite3
import traceback
from contextlib import redirect_stdout, redirect_stderr
import time
import math
from dataclasses import dataclass
from geoprob_pipe.results.construct_dataframes import (
    collect_df_beta_limit_state, collect_df_beta_scenario_rp, collect_df_beta_scenario_cp,
    collect_df_beta_scenario_final)
from geoprob_pipe.results.alphas_and_physical_values import (
    collect_stochast_values, calculate_derived_values)
import logging
from pandas import DataFrame

if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe
    from geoprob_pipe.calculations.systems.base_objects\
        .base_system_build import BaseSystemBuilder
    from geoprob_pipe.utils.validation_messages import ValidationMessages


logger = logging.getLogger("geoprob-pipe")


_BUILDER: BaseSystemBuilder
_MODEL: str


@dataclass
class CalcResult:
    """
    Dataclass om de resultaten te verzamelen vanuit de calculation.
    Bevat de volgende attributen:
    Dataframe: df_limit_state bevat resultaten van de afzonderlijke grenstoestandsfuncties (Z_u, Z_h en Z_p)
    Dataframe: df_scenario_rp bevat resultaten van max(Z_u, Z_h, Z_p).
    Dataframe: df_scenario_cp bevat resultaten van de combinatie van de afzonderlijke grenstoestandsfuncties
    Dataframe: df_scenario_final bevat resultaten op basis van beslisregels van de afzonderlijke berekeningsmethoden. Zie construct_dataframes.py voor meer details.
    Dataframe: df_stochast,
    Dataframe: df_derived,
    ValidationMessages: validation_message
    """
    df_limit_state: DataFrame
    df_scenario_rp: DataFrame
    df_scenario_cp: DataFrame
    df_scenario_final: DataFrame
    df_stochast: DataFrame
    df_derived: DataFrame
    validation_message: ValidationMessages


def _init_worker(geohydrologisch_model, geopackage_filepath,
                 to_run_vakken_ids):
    """ Initiator voor de worker, dit zorgt ervoor dat de tijdrovende
    stappen een keer per worker worden uitgevoerd en dan beschikbaar blijven
    voor iedere run.
    """
    global _BUILDER, _MODEL
    _MODEL = geohydrologisch_model
    _BUILDER = (
        CALCULATION_MAPPER[geohydrologisch_model]["system_builder"](
            geopackage_filepath=geopackage_filepath,
            to_run_vakken_ids=to_run_vakken_ids))


def _worker(row_unique: dict):
    """ De worker functie die op de parallelle rekenkernen wordt gedraaid.
    """
    log_buffer = StringIO()
    buffer_handler = logging.StreamHandler(log_buffer)
    buffer_handler.setLevel(logging.DEBUG)

    buffer_handler.setFormatter(logging.Formatter(
        "%(asctime)s [%(processName)s %(levelname)s] %(name)s: %(message)s",
        "%Y-%m-%d %H:%M:%S",
    ))

    root = logging.getLogger()
    prev_level = root.level
    root.setLevel(logging.DEBUG)
    root.addHandler(buffer_handler)
    logging.captureWarnings(True)

    # noinspection PyBroadException
    try:
        with redirect_stdout(log_buffer), redirect_stderr(log_buffer):
            # Build and run calculations
            calc = _BUILDER.build_instance(row_unique=row_unique)
            calc.run()

            # Collect results
            df_limit_state = collect_df_beta_limit_state(calc)
            df_scenario_rp = collect_df_beta_scenario_rp(calc)
            df_scenario_cp = collect_df_beta_scenario_cp(calc)
            df_scenario_final = collect_df_beta_scenario_final(calc)
            df_stochast = collect_stochast_values(calc, df_scenario_final=df_scenario_final)
            df_derived = calculate_derived_values(df_scenarios_final=df_scenario_final, geohydrologisch_model=_MODEL)
            df_scenario_rp = df_scenario_rp.drop(columns=["system_calculation"])
            df_scenario_cp = df_scenario_cp.drop(columns=["system_calculation"])
            df_scenario_final = df_scenario_final.drop(columns=["system_calculation"])

            # Return results (without calculation object)
            return CalcResult(
                df_limit_state=df_limit_state, df_scenario_rp=df_scenario_rp, df_scenario_cp=df_scenario_cp,
                df_scenario_final=df_scenario_final, df_stochast=df_stochast, df_derived=df_derived,
                validation_message=calc.validation_messages
            ), None, None

    except Exception:
        tb = traceback.format_exc()
        log_buffer.write(tb)
        buffer_handler.flush()
        return None, log_buffer.getvalue(), row_unique
    finally:
        # Handler altijd verwijderen
        logging.captureWarnings(False)
        root.removeHandler(buffer_handler)
        root.setLevel(prev_level)
        buffer_handler.close()


def build_and_run_system_calculations(geoprob_pipe: GeoProbPipe) -> List[CalcResult]:
    """ In deze functie worden de parameters voor de berekeningen verzamelt,
    aan de workers gegeven en vervolgens de resultaten verzameld.
    """
    geohydrologisch_model = geoprob_pipe.input_data.geohydrologisch_model
    geopackage_filepath = (
        geoprob_pipe.input_data.app_settings.geopackage_filepath)
    to_run_vakken_ids = geoprob_pipe.input_data.app_settings.to_run_vakken_ids
    system_builder: BaseSystemBuilder = (
        CALCULATION_MAPPER[geohydrologisch_model]['system_builder'](
            geopackage_filepath=geopackage_filepath,
            to_run_vakken_ids=to_run_vakken_ids))

    logger.info("Now building and running calculations...")
    df_unique_combos = system_builder.setup_iteration_df()
    # Bepaal de parameters voor de multiprocessing setup en de logger
    n_threads: int = cpu_count() - 1
    n_calc_totaal: int = len(df_unique_combos)
    # Minimaal 5 berekeningen per chunk en grootte van chunk beperken
    # zodat er gelogd kan worden.
    chunk_size: int = max(math.ceil(n_calc_totaal / (n_threads * 10)), 5)
    logger.info(
        f"Running {n_calc_totaal} calculations in chunks of {chunk_size}"
        f" with {n_threads} parallel threads.")
    char_len_total = str(n_calc_totaal).__len__()
    logger.info(
        f"Progress: {0:>{char_len_total}} / {n_calc_totaal} calculations.")

    # Dicts zijn gemakkelijker te pickelen en daardoor sneller te
    # verwerken dan pandas series.
    rows = [dict(zip(df_unique_combos.columns, r))
            for r in df_unique_combos.itertuples(index=False, name=None)]

    last_report = time.time()
    done = 0
    results: List[CalcResult] = []
    pool_size = max(min(math.floor(n_calc_totaal / chunk_size), n_threads), 1)

    # Multiprocessing setup
    error_rows = []
    with Pool(processes=pool_size, initializer=_init_worker, initargs=(
            geohydrologisch_model, geopackage_filepath, to_run_vakken_ids)) as pool:

        for res, error_logs, row in pool.imap_unordered(_worker, rows, chunksize=chunk_size):
            if isinstance(res, CalcResult):
                results.append(res)
            if isinstance(error_logs, str):
                error_rows.append({
                    "uittredepunt_id": row["uittredepunt_id"],
                    "ondergrondscenario_naam": row["ondergrondscenario_naam"],
                    "vak_id": row["vak_id"],
                    "error_logs": error_logs,
                })
            done += 1

            # Alleen kijken of er gelogd moet worden bij de laatste
            # berekening die uit de chunk komt.
            if done % chunk_size != 0:
                continue

            # Alleen loggen wanneer 30 seconden is gepasseerd
            now = time.time()
            if now - last_report < 30.0:
                continue

            # Log
            error_count_append = ""
            if error_rows.__len__() > 0:
                error_count_append = f" (of which {error_rows.__len__()} failed calculations)"
            logger.info(f"Progress: {done:>{char_len_total}} / {n_calc_totaal} calculations{error_count_append}.")
            last_report = now

    # Push errors to database (if any)
    conn = sqlite3.connect(geoprob_pipe.input_data.app_settings.geopackage_filepath)
    table_name = "calculation_error_logs"
    if error_rows.__len__() > 0:
        df_errors = DataFrame(data=error_rows)
        df_errors.to_sql(table_name, conn, if_exists="replace", index=False)
        conn.close()
        logger.error(f"There are {error_rows.__len__()} failed calculations. Error logs are stored inside the "
                     f"GeoPacakge in table '{table_name}'.")
    else:
        # Remove old table (if exists)
        cur = conn.cursor()
        cur.execute(f"DROP TABLE IF EXISTS {table_name};")
        conn.commit()
        conn.close()

    return results



from __future__ import annotations
from geoprob_pipe.utils.validation_messages import BColors
from InquirerPy import inquirer
from typing import TYPE_CHECKING
from geopandas import read_file
import sqlite3
from geoprob_pipe.calculations.systems.mappers.calculation_mapper import CALCULATION_MAPPER
from pandas import DataFrame
if TYPE_CHECKING:
    from geoprob_pipe.cmd_app.cmd import ApplicationSettings


def created_model(app_settings: ApplicationSettings) -> bool:
    df: DataFrame = read_file(app_settings.geopackage_filepath, layer="geoprob_pipe_metadata")

    # Check if no model specified yet
    metadata_types = [str(value) for value in df['metadata_type'].values.tolist()]
    if "geohydrologisch_model" not in metadata_types:
        specify_model_to_use(app_settings, update_record=False)
        return True

    # Check if specified model is legal
    current_specified_model = df[df['metadata_type'] == "geohydrologisch_model"]["values"].iloc[0]
    if current_specified_model not in CALCULATION_MAPPER.keys():
        specify_model_to_use(app_settings, update_record=True)
        return True

    # Specified model is legal
    print(BColors.OKBLUE, f"✔  Geohydrologisch model al ingesteld.", BColors.ENDC)
    return True


def specify_model_to_use(app_settings: ApplicationSettings, update_record: bool = False):
    model_labels = [value["label"] for key, value in CALCULATION_MAPPER.items()]

    choice = inquirer.select(
        message="Welk geohydrologisch model wil je gebruiken? De invoer parameters variëren per model.",
        choices=model_labels,
        default=model_labels[0],
    ).execute()

    # Push record
    conn = sqlite3.connect(app_settings.geopackage_filepath)
    cursor = conn.cursor()
    choice_formatted: str = choice
    choice_formatted = choice_formatted.replace(" ", "").lower()
    sql_statement = f"""
    INSERT INTO geoprob_pipe_metadata 
    ('metadata_type', 'values') VALUES ('geohydrologisch_model', '{choice_formatted}')"""
    if update_record:
        sql_statement = f"""
        UPDATE geoprob_pipe_metadata 
        SET 'values' = '{choice_formatted}' 
        WHERE metadata_type = 'geohydrologisch_model'"""
    cursor.execute(sql_statement)
    conn.commit()
    conn.close()

    print(BColors.OKBLUE, f"✅  Geohydrologisch model ingesteld.", BColors.ENDC)



import typer
from rich.console import Console
import os
from geoprob_pipe.utils import clear_terminal
from datetime import datetime
from rich.panel import Panel
from geoprob_pipe.cmd_app.questionnaire import start_questionnaire
from typing import Optional, List, Dict
from geoprob_pipe.cmd_app.utils.misc import get_geoprob_pipe_version_number
from geoprob_pipe.utils.loggers import setup_base_logging
from importlib.metadata import distributions

app = typer.Typer(help="GeoProb-Pipe - CLI applicatie voor probabilistische piping berekeningen.", add_completion=False)


class ApplicationSettings:

    def __init__(self):
        self.workspace_dir: Optional[str] = None
        self.geopackage_filename: Optional[str] = None
        self.datetime_stamp: str = datetime.now().strftime("%Y-%m-%d_%H%M%S")
        self.to_run = "all"
        # -> or vakken:1,2,3,4,5
        self.debug: bool = os.getenv("GEOPROB_DEBUG") == "1"

    @property
    def geopackage_filepath(self) -> str:
        return os.path.join(self.workspace_dir, self.geopackage_filename)

    @property
    def hrd_dir(self):
        path_to_hrd_dir = os.path.join(self.workspace_dir, "hrd_files")
        os.makedirs(path_to_hrd_dir, exist_ok=True)
        return path_to_hrd_dir

    @property
    def hrd_file_path(self) -> str:
        for file in os.listdir(self.hrd_dir):
            filename = os.fsdecode(file)
            if filename.endswith(".config.sqlite"):
                continue
            if filename.endswith("hlcd.sqlite"):
                continue
            return os.path.join(self.hrd_dir, filename)
        raise ValueError

    @property
    def ahn_filepath(self) -> str:
        return os.path.join(self.workspace_dir, "ahn", "ahn.tif")

    @property
    def to_run_vakken_ids(self) -> Optional[List[int]]:
        if self.to_run == "all":
            return None
        vak_ids_str: List[str] = self.to_run.replace("vakken:", "").split(sep=",")
        return [int(vak_id_str) for vak_id_str in vak_ids_str]


def _raise_if_multiple_installations():
    installed_distributions: List[Dict[str, str]] = []
    for dist in distributions():
        if "geoprob" in dist.metadata["Name"].lower():
            installed_distributions.append({
                "NAME": dist.metadata["Name"],
                "VERSION": dist.version,
                "LOCATION": dist.locate_file(""),
            })
    if installed_distributions.__len__() > 1:
        raise RuntimeError(
            f"Possibly multiple GeoProb-Pipe installations in your Python environment. \n\n"
            f"Please uninstall GeoProb-Pipe first with the commands `pip uninstall geoprob-pipe geoprob_pipe`. \n"
            f"Run this command until Python can't find anymore GeoProb-Pipe installations. \n"
            f"Then re-install geoprob-pipe with the command `pip install geoprob-pipe`. \n"
            f"For your information, the following installations were found: \n"
            f"{installed_distributions}")


def startup_geoprob_pipe():
    """ Starts up the GeoProb-Pipe console application. """

    _raise_if_multiple_installations()

    app_settings = ApplicationSettings()

    setup_base_logging()

    debug_label: str = ""
    if app_settings.debug:
        debug_label = f", DEBUG=TRUE"

    clear_terminal()
    console = Console()
    console.print(Panel(
        """
Welkom bij GeoProb-Pipe! Deze applicatie voert probabilistische pipingberekeningen uit met de uittredepuntenmethode en
een geohydrologisch model naar keuze (zoals model4a). GeoProb-Pipe maakt gebruik van de probabilistische bibliotheek
van Deltares, die onder de motorkap de PTK-tool aanstuurt. Met de onderstaande interactieve vragenmodule neemt
GeoProb-Pipe je stap voor stap mee door het opzetten van de invoer en het uitvoeren van de berekeningen.
""",
        title=f"GeoProb-Pipe ({get_geoprob_pipe_version_number()}{debug_label})".upper(),
        title_align="left",
        border_style="bright_blue",
        padding=(0, 2)))

    start_questionnaire(app_settings=app_settings)


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """ Default entry point for `geoprob-pipe`. Runs when no subcommand is specified. """
    if ctx.invoked_subcommand is None:
        startup_geoprob_pipe()


@app.command()
def debug():
    """ Start GeoProb-Pipe in debug mode. """
    os.environ["GEOPROB_DEBUG"] = "1"
    startup_geoprob_pipe()



from geoprob_pipe.calculations.limit_states.piping_lm import (
    limit_state_model4a)


# noinspection PyPep8Naming
def calc_Z_u(
        L_intrede: float, L_but: float, L_bit: float, L_achterland: float,
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        top_zand: float, kD_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float, c_voorland: float, c_achterland: float,
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float, d70_m: float,
        gamma_korrel: float, v: float, theta: float, eta: float, g: float,
        gamma_water: float
) -> float:
    r"""Grenstoestandfunctie voor opbarsten (uplift).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor opbarsten
    """

    return limit_state_model4a(
        L_intrede=L_intrede, L_but=L_but, L_bit=L_bit,
        L_achterland=L_achterland, buitenwaterstand=buitenwaterstand,
        polderpeil=polderpeil, mv_exit=mv_exit, top_zand=top_zand,
        kD_wvp=kD_wvp, D_wvp=D_wvp, d70=d70,
        gamma_sat_deklaag=gamma_sat_deklaag, c_voorland=c_voorland,
        c_achterland=c_achterland, modelfactor_u=modelfactor_u,
        modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d,
        modelfactor_aniso=modelfactor_aniso, modelfactor_ml=modelfactor_ml,
        i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, d70_m=d70_m,
        gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
        gamma_water=gamma_water)[0]


# noinspection PyPep8Naming
def calc_Z_h(
        L_intrede: float, L_but: float, L_bit: float, L_achterland: float,
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        top_zand: float, kD_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float, c_voorland: float, c_achterland: float,
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float, d70_m: float,
        gamma_korrel: float, v: float, theta: float, eta: float, g: float,
        gamma_water: float
) -> float:
    r""" Wrapper over de grenstoestandfunctie voor heave zodat deze bruikbaar
    is voor de Probabilistic Library.

    Returns:
        float: Z waarde van de grenstoestandfunctie voor heave
    """
    return limit_state_model4a(
        L_intrede=L_intrede, L_but=L_but, L_bit=L_bit,
        L_achterland=L_achterland, buitenwaterstand=buitenwaterstand,
        polderpeil=polderpeil, mv_exit=mv_exit, top_zand=top_zand,
        kD_wvp=kD_wvp, D_wvp=D_wvp, d70=d70,
        gamma_sat_deklaag=gamma_sat_deklaag, c_voorland=c_voorland,
        c_achterland=c_achterland, modelfactor_u=modelfactor_u,
        modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d,
        modelfactor_aniso=modelfactor_aniso, modelfactor_ml=modelfactor_ml,
        i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, d70_m=d70_m,
        gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
        gamma_water=gamma_water)[1]


# noinspection PyPep8Naming
def calc_Z_p(
        L_intrede: float, L_but: float, L_bit: float, L_achterland: float,
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        top_zand: float, kD_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float, c_voorland: float, c_achterland: float,
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float, d70_m: float,
        gamma_korrel: float, v: float, theta: float, eta: float, g: float,
        gamma_water: float
) -> float:
    r"""Grenstoestandfunctie voor het mechanisme terugschrijdende erosie (piping).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor piping
    """
    return limit_state_model4a(
        L_intrede=L_intrede, L_but=L_but, L_bit=L_bit,
        L_achterland=L_achterland, buitenwaterstand=buitenwaterstand,
        polderpeil=polderpeil, mv_exit=mv_exit, top_zand=top_zand,
        kD_wvp=kD_wvp, D_wvp=D_wvp, d70=d70,
        gamma_sat_deklaag=gamma_sat_deklaag, c_voorland=c_voorland,
        c_achterland=c_achterland, modelfactor_u=modelfactor_u,
        modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d,
        modelfactor_aniso=modelfactor_aniso, modelfactor_ml=modelfactor_ml,
        i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, d70_m=d70_m,
        gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
        gamma_water=gamma_water)[2]


# noinspection PyPep8Naming
def calc_Z_project(
        L_intrede: float, L_but: float, L_bit: float, L_achterland: float,
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        top_zand: float, kD_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float, c_voorland: float, c_achterland: float,
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float, d70_m: float,
        gamma_korrel: float, v: float, theta: float, eta: float, g: float,
        gamma_water: float
) -> float:
    r"""Grenstoestandfunctie voor het mechanisme piping. Deze grenstoestandfunctie bepaalt de Z waarde 
    op basis van de afzonderlijke Z waarden van de mechanismen uplift, heave en terugschrijdende erosie door
    max(Z_u, Z_h, Z_p).

    Returns:
        float: Z waarde van de grenstoestandfunctie voor piping
    """
    return limit_state_model4a(
        L_intrede=L_intrede, L_but=L_but, L_bit=L_bit,
        L_achterland=L_achterland, buitenwaterstand=buitenwaterstand,
        polderpeil=polderpeil, mv_exit=mv_exit, top_zand=top_zand,
        kD_wvp=kD_wvp, D_wvp=D_wvp, d70=d70,
        gamma_sat_deklaag=gamma_sat_deklaag, c_voorland=c_voorland,
        c_achterland=c_achterland, modelfactor_u=modelfactor_u,
        modelfactor_h=modelfactor_h, modelfactor_p=modelfactor_p,
        modelfactor_ff=modelfactor_ff, modelfactor_3d=modelfactor_3d,
        modelfactor_aniso=modelfactor_aniso, modelfactor_ml=modelfactor_ml,
        i_c_h=i_c_h, r_c_deklaag=r_c_deklaag, d70_m=d70_m,
        gamma_korrel=gamma_korrel, v=v, theta=theta, eta=eta, g=g,
        gamma_water=gamma_water)[3]


MODEL_NAMES = {
    calc_Z_h.__name__: "Heave",
    calc_Z_u.__name__: "Uplift",
    calc_Z_p.__name__: "Piping",
}



from __future__ import annotations
import os
# import matplotlib.pyplot as plt
from typing import TYPE_CHECKING, List
import plotly.graph_objects as go
from datetime import datetime
# from geopandas import GeoDataFrame, read_file
# from probabilistic_library import FragilityValue
import pydra_core as pydra
from pandas import Series, concat, DataFrame, read_sql
from geoprob_pipe.cmd_app.parameter_input.expand_input_tables import run_expand_input_tables
import sqlite3

if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


class GraphHFreqSingleInteractive:

    def __init__(self, geoprob_pipe: GeoProbPipe, export: bool = False):

        # Check if there are fragility curves referenced:
        conn = sqlite3.connect(geoprob_pipe.input_data.app_settings.geopackage_filepath)
        df_parameter_invoer = read_sql("SELECT * FROM parameter_invoer;", conn)
        fragility_values_refs = df_parameter_invoer['fragility_values_ref'].unique()
        conn.close()
        if (fragility_values_refs.__len__() == 0 or
                (fragility_values_refs.__len__() == 1 and fragility_values_refs[0] == '')):
            return  # No graphs needed

        # Helper parameters
        self.max_level: float = -999
        self.min_level: float = 999
        self.max_p: float = 0.0
        self.min_p: float = 1.0

        # Class-wide used
        self.geoprob_pipe = geoprob_pipe
        self.df_parameter_input_expanded: DataFrame = run_expand_input_tables(
            geopackage_filepath=self.geoprob_pipe.input_data.app_settings.geopackage_filepath, add_frag_ref=True)
        self.df_fragility_ref_data: DataFrame = self._collect_fragility_ref_data()
        if self.df_fragility_ref_data.__len__() == 0:
            return  # No fragility values, then no graphs needed

        # Logic
        self.fig = go.Figure()
        self._add_ondergrens()
        self._add_signaleringswaarde()
        self._add_dummy_physical_value_legend_marker()
        self._add_overschrijdingsfrequentielijnen()
        self._add_physical_values()
        self._update_layout()
        self._optionally_export(export=export)

    def _collect_fragility_ref_data(self) -> DataFrame:

        # Transform parameter_input dictionary to columns (added columns will be: fragility_values and fragility_ref)
        df = self.df_parameter_input_expanded
        df = df[df["parameter_name"] == "buitenwaterstand"]
        df = concat([df.drop(columns=['parameter_input']), df['parameter_input'].apply(Series)], axis=1)
        # print(f"{df.__len__()=}")
        # print(f"{df.columns=}")
        # print(f"{df=}")

        # If fragility_values_ref not in columns, then there are no fragility values needed
        if "fragility_values_ref" not in df.columns:
            return DataFrame(
                data=[],
                columns=["uittredepunt_id", "fragility_values", "uittredepunt_ids_multiline", "frequency_line"]
            )

        # Group uittredepunt ids and fragility values
        df_result = (
            df.groupby('fragility_values_ref').agg({
                'uittredepunt_id': lambda x: ', '.join(map(str, x)),  # concatenate IDs
                'fragility_values': 'first'  # keep the first list (since all are equal per ref)
            }).reset_index())
        df_result = df_result.rename(columns={"uittredepunt_id": "uittredepunt_ids"})

        # Create multiline item for uittredepunten ids
        df_result['uittredepunt_ids_multiline'] = ""
        df_result['frequency_line'] = ""
        rows = []
        for _, row in df_result.iterrows():

            # Uittredepunt ids multiline
            arr = row['uittredepunt_ids'].split(", ")
            lines = [', '.join(arr[i:i + 5]) for i in range(0, len(arr), 5)]
            row['uittredepunt_ids_multiline'] = '<br>'.join(lines)

            # Create frequency line
            fragility_values = row['fragility_values']
            levels = [item.x for item in fragility_values]
            exceedance_frequencies = [item.probability_of_failure for item in fragility_values]
            # noinspection PyUnresolvedReferences
            row['frequency_line'] = pydra.core.datamodels.frequency_line.FrequencyLine(
                level=levels, exceedance_frequency=exceedance_frequencies)
            rows.append(row)
        return DataFrame(rows)

    def _collect_used_fragility_refs(self) -> List[str]:
        df = self.df_parameter_input_expanded
        df = df[df["parameter_name"] == "buitenwaterstand"]

        # Transform parameter_input dictionary to columns (added columns will be: fragility_values and fragility_ref)
        df = concat([df.drop(columns=['parameter_input']), df['parameter_input'].apply(Series)], axis=1)

        return df['fragility_values_ref'].unique()

    def _add_ondergrens(self):
        ondergrens = 1 / self.geoprob_pipe.input_data.traject_normering.ondergrens
        self.fig.add_trace(go.Scatter(
            x=[-100, 999],
            y=[ondergrens, ondergrens],
            mode='lines',
            name=f"Ondergrens (1/{self.geoprob_pipe.input_data.traject_normering.ondergrens:,} jaren)",
            line=dict(color='black', width=3),
            showlegend=True,
        ))

    def _add_signaleringswaarde(self):
        signaleringswaarde = 1 / self.geoprob_pipe.input_data.traject_normering.signaleringswaarde
        self.fig.add_trace(go.Scatter(
            x=[-100, 999],
            y=[signaleringswaarde, signaleringswaarde],
            mode='lines',
            name=f"Signaleringswaarde (1/{self.geoprob_pipe.input_data.traject_normering.signaleringswaarde:,} jaren)",
            line=dict(dash='dash', color='black', width=3),
            showlegend=True,
        ))

    def _add_dummy_physical_value_legend_marker(self):
        # Dummy legend marker
        self.fig.add_trace(go.Scatter(
            x=[-99], y=[1], mode='markers', name='Physical values', showlegend=True,
            marker=dict(color='LightSkyBlue', size=10, line=dict(color='black', width=1))))

    def _get_fragility_values_ref_data(self, ref: str):
        df = self.df_fragility_ref_data
        df = df[df['fragility_values_ref'] == ref]
        return (df['fragility_values'].iloc[0],
                df['frequency_line'].iloc[0],
                [int(item) for item in df['uittredepunt_ids'].iloc[0].replace(" ", "").split(",")],
                df['uittredepunt_ids_multiline'].iloc[0])

    def _add_overschrijdingsfrequentielijnen(self):

        for index, fragility_values_ref in enumerate(self.df_fragility_ref_data["fragility_values_ref"].unique()):

            # Collect data for the graph
            fragility_values, _, uittredepunten, uittredepunten_multiline = self._get_fragility_values_ref_data(
                ref=fragility_values_ref)
            levels = [item.x for item in fragility_values]
            self.max_level = max(self.max_level, max(levels))
            self.min_level = min(self.min_level, min(levels))
            freqs = [item.probability_of_failure for item in fragility_values]
            self.max_p = max(self.max_p, max(freqs))
            self.min_p = min(self.min_p, min(freqs))

            # Only first overschrijdingsfrequentielijn should be visible at first
            visible = 'legendonly'
            if index == 0:
                visible = True

            # Add lines
            legend_name = f"{fragility_values_ref}<br>uittredepunten: <br>{uittredepunten_multiline}"
            self.fig.add_trace(go.Scatter(
                x=levels, y=freqs, mode='lines', name=legend_name, legendgroup=legend_name, visible=visible,
                line=dict(dash='dash', color='blue', width=1.5), showlegend=True))

    def _add_physical_values(self):

        for index, fragility_values_ref in enumerate(self.df_fragility_ref_data["fragility_values_ref"].unique()):

            # Collect data for the graph
            _, freq_line, uittredepunten, uittredepunten_multiline = self._get_fragility_values_ref_data(
                ref=fragility_values_ref)

            # Only first overschrijdingsfrequentielijn should be visible at first
            visible = 'legendonly'
            if index == 0:
                visible = True

            # Get physical values
            legend_name = f"{fragility_values_ref}<br>uittredepunten: <br>{uittredepunten_multiline}"
            df = self.geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(
                filter_deterministic=False, filter_derived=True)
            df = df[df['variable'] == 'buitenwaterstand']
            df = df[df['uittredepunt_id'].isin(uittredepunten)]

            # Scenario may have no uittredepunt connections, then no physical values need to be added
            if df.__len__() == 0:
                return

            levels = df['physical_value'].values
            self.max_level = max(self.max_level, levels.max())
            self.min_level = min(self.min_level, levels.min())
            frequencies = [freq_line.interpolate_level(level) for level in levels]
            self.max_p = max(self.max_p, max(frequencies))
            self.min_p = min(self.min_p, min(frequencies))

            # Add markers
            self.fig.add_trace(go.Scatter(
                x=levels, y=frequencies, mode='markers', visible=visible, showlegend=False, legendgroup=legend_name,
                marker=dict(color='LightSkyBlue', size=10, line=dict(color='black', width=1))))

    def _yticks(self):
        min_range = int(f"{self.min_p:.0e}".split("e")[1])
        e_values = list(range(0, min_range - 15, -1))
        y_ticks = [10 ** e_value for e_value in e_values]
        y_ticks_text = [f"10<sup>{e_value}</sup>" for e_value in e_values]
        return y_ticks, y_ticks_text

    def _update_layout(self):
        length_range_xaxis = self.max_level - self.min_level
        xaxis_add = length_range_xaxis * 0.05
        y_ticks, y_ticks_text = self._yticks()
        self.fig.update_layout(
            title=f"<b>Overschrijdingsfrequentielijnen voor alle HydraNL locaties</b><br>"
                  f"<sup>Traject {self.geoprob_pipe.input_data.traject_normering.traject_id}</sup>",
            xaxis=dict(
                title=f"Waterstand (m+NAP)", type='linear', showgrid=True, gridwidth=0.5, gridcolor="gray",
                range=[self.min_level - xaxis_add, self.max_level + xaxis_add]),
            yaxis=dict(
                title=f"Overschrijdingsfrequentie (log-schaal)", type='log', showgrid=True, tickformat=".0e",
                gridwidth=1.0, tickvals=y_ticks, ticktext=y_ticks_text, tickmode='array', gridcolor="gray",
                minor=dict(
                    showgrid=True, dtick="D1", gridwidth=0.5, gridcolor='rgb(199, 197, 193)')))

    def _optionally_export(self, export: bool = False, add_timestamp: bool = False):
        if not export:
            return
        export_dir = self.geoprob_pipe.visualizations.graphs.export_dir
        os.makedirs(export_dir, exist_ok=True)
        timestamp_str = ""
        if add_timestamp:
            timestamp = datetime.now().strftime("%Y-%m-%d_%H%M")
            timestamp_str = f"{timestamp}_"
        self.fig.write_html(os.path.join(export_dir, f"{timestamp_str}hfreq.html"), include_plotlyjs='cdn')
        # if self.geoprob_pipe.software_requirements.chrome_is_installed:
        #     self.fig.write_image(os.path.join(export_dir, f"{timestamp_str}hfreq.png"), format="png")
        # Note CP: Export van PNG uitgezet. Meerwaarde beperkt denk ik aangezien je de HTML al hebt.



r"""Module met functies voor het berekenen van fysische componenten van
piping en uplift. Dit betreft onder andere de dikte van de deklaag,
het niveau bij het uittredepunt en de kwelweglengte.
"""

import math
from geoprob_pipe.calculations.physical_components.model4a import Model4a


def calc_d_deklaag(
        mv_exit: float,
        top_zand: float
) -> float:
    r"""Berekening deklaagdikte ter plaatse van het uittredepunt,
    de minimale dikte van de deklaag is 0.1 m omdat negatieve deklaagdiktes
    niet mogelijk zijn. Dit uitgangspunt is gekozen omdat ook bij een zeer
    dunne deklaag nog enige reductie van het verval verwacht mag worden.

    Args:
        mv_exit (float): Bodemhoogte ter plaatse van Uittredepunten [m+NAP]
        top_zand (float): Geschematiseerde top van het vak [m+NAP]

    Returns:
        float: deklaagdikte [m]
    """
    return max(mv_exit - top_zand, 0.1)


def calc_h_exit(
        polderpeil: float,
        mv_exit: float
) -> float:
    r"""Berekening van het niveau van het uittredepunt op basis van polderpeil
    of maaiveldniveau. Functie geeft de maximale waarde van polderpeil en
    mv_exit terug. Dit is de benedenstroomse randvoorwaarde voor het verval
    in pipingberekeningen.

    Args:
        polderpeil (float): polderpeil [m+NAP]
        mv_exit (float): maaiveldniveau van uittredepunt [m+NAP]

    Returns:
        float: niveau bij het uittredepunt in m+NAP
    """
    return max(polderpeil, mv_exit)


# noinspection PyPep8Naming
def calc_lengte_voorland(
        L_intrede: float,
        L_but: float
) -> float:
    r""" Berekent de geometrische voorlandlengte in [m] op basis van
    afstanden ten opzichte van een uittredepunt. In de
    pre-processing tool worden :math:`L_{intrede}` en :math:`L_{but}`
    als geografische lijnobjecten gedefinieerd. De kortste afstand tussen
    deze objecten is invoer voor deze functie.

    Args:
        L_intrede (float): afstand van uittredepunten tot een (denkbeeldige)
        intredelijn [m].
        L_but (float): afstand van uittredepunten tot buitenteenlijn [m].

    Returns:
        float: geometrische voorlandlengte [m]
    """
    return abs(L_intrede - L_but)


# noinspection PyPep8Naming
def calc_lambda_achterland(
        kD_wvp: float,
        c_achterland: float
) -> float:
    r"""Berekent de spreidingslengte van het achterland in [m].

    .. math::

        \lambda = \sqrt{kDc}

    Args:
        kD_wvp (float): Transmissiviteit van het watervoerende pakket [m²/dag]
        c_achterland (float): Weerstand van de deklaag in het achterland [dag]

    Returns:
        float: spreidingslengte van het achterland [m]
    """
    return (kD_wvp * c_achterland) ** (1 / 2)


# noinspection PyPep8Naming
# TODO: functie samenvoegen met calc_lambda_achterland?
def calc_lambda_voorland(
        kD_wvp: float,
        c_voorland: float
) -> float:
    r"""Berekent de spreidingslengte van het achterland in [m].

    .. math::

        \lambda = \sqrt{kDc}

    Args:
        kD_wvp (float): Transmissiviteit van het watervoerende pakket [m²/dag]
        c_voorland (float): Weerstand van de deklaag in het voorland [dag]

    Returns:
        float: spreidingslengte van het voorland [m]
    """
    return (kD_wvp * c_voorland) ** (1 / 2)


def calc_dh_red(
        buitenwaterstand: float,
        h_exit: float,
        r_c_deklaag: float,
        d_deklaag: float
) -> float:
    r"""Berekening van het gereduceerde verval over de waterkering.

    .. math::

        \Delta h_{red} = h_{buitenwaterstand} - h_{exit} - r_{c, deklaag}
        \cdot d_{deklaag}

    Args:
        buitenwaterstand (float): buitenwaterstand [m+NAP]
        h_exit (float): Benedenstroomse randvoorwaarde verval [m+NAP]
        r_c_deklaag (float): Reductie constante van het verval over de
        deklaag [-]
        d_deklaag (float): deklaagdikte in m

    Returns:
        float: gereduceerd verval [m]
    """
    return buitenwaterstand - h_exit - r_c_deklaag * d_deklaag


# noinspection PyPep8Naming
def calc_W_achterland(
        lambda_achterland: float,
        L_achterland: float
) -> float:
    r"""Berekent de geohydrologische weerstand van het achterland in [m].

    .. math::

        W = \lambda tanh(\frac{L}{\lambda})

    Args:
        lambda_achterland (float): de spreidingslengte van het achterland [m]
        L_achterland (float): afstand van uittredepunten tot
        achterlandlengte [m]

    Returns:
        float: geohydrologische weerstand van het achterland [m]
    """
    return lambda_achterland * math.tanh(L_achterland / lambda_achterland)


# noinspection PyPep8Naming
# TODO: functie samenvoegen met calc_W_achterland?
def calc_W_voorland(
        lambda_voorland: float,
        L_voorland: float
) -> float:
    r""" Berekent de geohydrologische weerstand van het voorland in [m].
    Dit wordt ook wel de effectieve voorlandlengte genoemd.

    .. math::

        W = \lambda tanh(\frac{L}{\lambda})

    Args:
        lambda_voorland (float): de spreidingslengte van het voorland [m]
        L_voorland (float): Geometrische voorlandlengte [m]

    Returns:
        float: geohydrologische weerstand van het voorland [m]
    """
    return lambda_voorland * math.tanh(L_voorland / lambda_voorland)


# noinspection PyPep8Naming
def calc_L_kwelweg(
        L_but: float,
        W_voorland: float
) -> float:
    r"""Berekent de kwelweglengte in [m].
    De kwelweglengte is de som van de afstand van het uittredepunt tot de
    buitenteenlijn en de effectieve voorlandlengte van het voorland.
    De onzekerheid in de kwelweglengte zit in de effectieve voorlandlengte.

    Args:
        L_but (float): afstand van uittredepunten tot buitenteenlijn [m]
        W_voorland (float): geohydrologische weerstand van het voorland [m]

    Returns:
        float: kwelweglengte [m]
    """

    return W_voorland + L_but


def calc_dphi_c_u(
        d_deklaag: float,
        gamma_sat_deklaag: float,
        gamma_water: float
) -> float:
    r"""Berekening grenspotentiaal ten opzichte van maaiveldniveau in [m].

    .. math::

        \Delta \phi_{c, u} = \frac{d_{deklaag} \cdot (\gamma_{sat, deklaag}
        - \gamma_{w})}{\gamma_{w}}

    Args:
        d_deklaag (float): Dikte van de cohesieve deklaag [m]
        gamma_sat_deklaag (float): verzadigd volumegewicht van de
        deklaag [kN/m³]
        gamma_water (float): volumegewicht van water [kN/m³]

    Returns:
        float: grenspotentiaal ten opzichte van maaiveldniveau [m]
    """
    return d_deklaag * (gamma_sat_deklaag - gamma_water) / gamma_water


def calc_i_exit(
        phi_exit: float,
        h_exit: float,
        d_deklaag: float
) -> float:
    r""" Berekening van de optredende heave gradiënt. De heave gradient is
    het stijghoogteverschil over de deklaag gedeeld door de deklaagdikte.

    .. math::

        i_{exit} = \frac{(\phi_{exit} - h_{exit})}{d_{deklaag}}

    Args:
        phi_exit (float): stijghoogte in het watervoerende zandpakket ter
        plaatse van uittredepunt in m+NAP
        h_exit (float): niveau bij het uittredepunt [m+NAP]
        d_deklaag (float): deklaagdikte [m]

    Returns:
        float: heave gradient in [-]
    """
    return (phi_exit - h_exit) / d_deklaag


# noinspection PyPep8Naming
# TODO: deze wrapper functie wordt gebruikt in heave_icw_model4a.py
# en uplift_icw_model4a.py
# check of deze limit_state functies ook daadwerkelijk gebruikt worden
# in de berekeningen.
# zo niet, verwijder deze functies of roep de model4a klasse direct aan
# in de limit_state functies.
def calc_r_exit_model4a(
        kD_wvp: float,
        D_wvp: float,
        c_voorland: float,
        c_achterland: float,
        L_but: float,
        L_bit: float,
        L_achterland: float,
        L_voorland: float
) -> float:
    r"""Wrapper functie voor het berekenen van de dempingsfactor bij
    uittredepunten met behulp van Model4a. De functie gaat uit dat
    x = 0.0 bij de binnenteen ligt.  Dit betekent dat x_bit = 0.0
    en x_but negatief is.
    Uittredepunten moeten altijd binnendijks van de binnenteenlijn liggen.

    """
    model4a = Model4a(
        kD=kD_wvp,
        D=D_wvp,
        c1=c_voorland,
        c3=c_achterland,
        L1=L_voorland,
        L3=L_achterland,
        # x_but moet negatief zijn, x_bit is 0.0
        x_but=-1.0 * abs(L_but - L_bit),
        x_bit=0.0, )  # x_bit is 0.0
    # Bereken de respons bij het uittredepunt
    r_exit, _, _ = model4a.respons(L_bit)
    return r_exit


def calc_phi_exit(
        polderpeil: float,
        r_exit: float,
        buitenwaterstand: float
) -> float:  # Van respons naar potentiaal
    r"""Berekent de theoretische stijghoogte bij uittredepunten in [m+NAP].

    .. math::

        \phi_exit(x) = polderpeil + r(x) (buitenwaterstand - polderpeil)

    Args:
        polderpeil (float): Benedenstroomse randvoorwaarde verval [m+NAP]
        r_exit (float): Dempingsfactor bij uittredepunten [-]
        buitenwaterstand (float): buitenwaterstand [m+NAP]

    Returns:
        float: Theoretische stijghoogte bij uittredepunten [m+NAP]
    """
    return polderpeil + r_exit * (buitenwaterstand - polderpeil)


# noinspection PyPep8Naming
# TODO: functies toevoegen in docstring
def calc_dh_c(
        d70: float,
        D_wvp: float,
        kD_wvp: float,
        L_kwelweg: float,
        gamma_water: float,
        g: float,
        v: float,
        theta: float,
        eta: float,
        d70_m: float,
        gamma_korrel: float,
) -> float:
    r"""Berekening kritiek verval methode Sellmeijer inclusief
    berekeningsinstellingen

    .. math::

    \Delta H_{c} = F_{resistance} \cdot F_{scale} \cdot F_{geometry}
    \cdot L_{kwelweg}


    Args:
        d70 (float): 70% percentiel van de korrelgrootteverdeling [m]
        D_wvp (float): dikte van het watervoerende pakket [m]
        kD_wvp (float): transmissiviteit van het watervoerende pakket [m²/dag]
        L_kwelweg (float): kwelweglengte in meters
        gamma_water (float): volumegewicht van water [kN/m³]
        g (float): Zwaartekrachtversnelling [m/s2]
        v (float): kinematische viscositeit [m²/s]
        theta (float): rolweerstandshoek [graden]
        eta (float): coefficiënt van White [-]
        d70_m (float): gemiddelde d70 in kleine schaalproeven [m]
        gamma_korrel (float): (schijnbaar) volumegewicht van de zandkorrels
        onder water [kN/m³]

    Returns:
        float: kritiek verval [m]
    """
    # Omrekenen transmissiviteit naar doorlatendheid
    k_wvp_calc = kD_wvp / D_wvp  

    # Omrekenen doorlatendheid van m/d naar m/s
    k_wvp_calc_sec = k_wvp_calc / (24 * 3600)
    # Intrinsieke doorlatendheid
    k_intr = (v / g) * k_wvp_calc_sec

    # Berekening Fres
    # noinspection PyPep8Naming
    Fres = (
            eta
            * ((gamma_korrel - gamma_water) / gamma_water)
            * math.tan(theta * math.pi / 180.00)
    )

    # Berekening Fscale
    # noinspection PyPep8Naming
    Fscale = (pow(d70 / d70_m, 0.4) * d70_m
              / pow(k_intr * L_kwelweg, (1.0 / 3.0)))

    # Berekening F_geometry
    if D_wvp == L_kwelweg:
        # noinspection PyPep8Naming
        D_wvp = D_wvp - 0.001
    else:
        pass
    totdemacht = 0.04 + (0.28 / (pow(D_wvp / L_kwelweg, 2.8) - 1.0))
    # noinspection PyPep8Naming
    Fgeom = 0.91 * pow(D_wvp / L_kwelweg, totdemacht)

    return Fres * Fscale * Fgeom * L_kwelweg



"""Alternatieve implementatie van de grenstoestandsfuncties voor het
WBI-model, model 4a en MORIA. Alle grenstoestandsfuncties worden
gecombineerd in één functie per model. De fysische componenten worden
geïmporteerd vanuit de subpackage `physical_components.piping`. Deze
implementatie is bedoeld om de leesbaarheid en onderhoudbaarheid van
de code te verbeteren."""

from typing import Tuple

import geoprob_pipe.calculations.physical_components.piping as pc_piping


# noinspection PyPep8Naming
def limit_state_wbi(
        # Geometry parameters
        L_kwelweg: float,
        # Boundary condition parameters
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        top_zand: float, r_exit: float, k_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float,
        # Overige parameters
        gamma_water: float,
        # Constants  # TODO: Ombouwen tot globals
        d70_m: float, gamma_korrel: float, v: float, theta: float, eta: float, g: float,
) -> Tuple[float, ...]:
    """Grenstoestandsfuncties volgens het standaard WBI-model.

    :param L_kwelweg:
    :param buitenwaterstand:
    :param polderpeil:
    :param mv_exit:
    :param top_zand:
    :param r_exit:
    :param k_wvp:
    :param D_wvp:
    :param d70:
    :param gamma_sat_deklaag:
    :param gamma_water:
    :param modelfactor_u:
    :param modelfactor_h:
    :param modelfactor_p:
    :param modelfactor_ff: Model factor fijne fractie
    :param modelfactor_3d: Model factor 3D effecten
    :param modelfactor_aniso: Model factor anisotropie
    :param modelfactor_ml: Model factor meerlaagsheid zandpakket
    :param i_c_h:
    :param r_c_deklaag:
    :param d70_m:
    :param gamma_korrel:
    :param v:
    :param theta:
    :param eta:
    :param g:
    :return:
    """
    d_deklaag = pc_piping.calc_d_deklaag(mv_exit=mv_exit, top_zand=top_zand)
    phi_exit = pc_piping.calc_phi_exit(
        polderpeil=polderpeil, r_exit=r_exit,
        buitenwaterstand=buitenwaterstand)
    h_exit = pc_piping.calc_h_exit(polderpeil=polderpeil, mv_exit=mv_exit)
    dphi_c_u = pc_piping.calc_dphi_c_u(
        d_deklaag=d_deklaag, gamma_sat_deklaag=gamma_sat_deklaag,
        gamma_water=gamma_water)
    i_exit = pc_piping.calc_i_exit(
        phi_exit=phi_exit, h_exit=h_exit, d_deklaag=d_deklaag)
    dh_c = pc_piping.calc_dh_c(
        d70=d70, D_wvp=D_wvp, kD_wvp=k_wvp * D_wvp, L_kwelweg=L_kwelweg,
        gamma_water=gamma_water, g=g, v=v, theta=theta, eta=eta, d70_m=d70_m,
        gamma_korrel=gamma_korrel)
    dh_red = pc_piping.calc_dh_red(
        buitenwaterstand=buitenwaterstand, h_exit=h_exit,
        r_c_deklaag=r_c_deklaag, d_deklaag=d_deklaag)
    z_u = modelfactor_u * dphi_c_u - (phi_exit - h_exit)
    z_h = (modelfactor_h * i_c_h) - i_exit
    z_p = (modelfactor_p * modelfactor_ff * modelfactor_3d * modelfactor_aniso * modelfactor_ml * dh_c) - dh_red
    z_combin = max(z_u, z_h, z_p)
    return z_u, z_h, z_p, z_combin, h_exit, phi_exit, dphi_c_u, i_exit, dh_c, dh_red


# Define the input variables for the model4 limit state
# The variables are grouped into categories for clarity
# --geometry--
# L_intrede: float
# L_but: float
# L_bit: float
# L_achterland: float
# --Boundary conditions--
# buitenwaterstand: float
# polderpeil: float
# mv_exit: float
# --subsoil properties--
# top_zand: float
# kD_wvp: float
# D_wvp: float
# d70: float
# gamma_sat_deklaag: float
# c_voorland: float
# c_achterland: float
# --model properties--
# modelfactor_u: float
# modelfactor_h: float
# modelfactor_p: float
# modelfactor_ff: float model factor fijne fractie
# modelfactor_3d: float model factor 3D effecten
# modelfactor_aniso: float model factor anisotropie
# modelfactor_ml: float model factor meerlaagsheid zandpakket
# i_c_h: float
# r_c_deklaag: float
# d70_m: float
# gamma_korrel: float
# v: float
# theta: float
# eta: float
# --constants--
# g: float
# gamma_water: float


# noinspection PyPep8Naming
def limit_state_model4a(
        L_intrede: float, L_but: float, L_bit: float, L_achterland: float,
        buitenwaterstand: float, polderpeil: float, mv_exit: float,
        top_zand: float, kD_wvp: float, D_wvp: float, d70: float,
        gamma_sat_deklaag: float, c_voorland: float, c_achterland: float,
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float, modelfactor_aniso: float,
        modelfactor_ml: float, i_c_h: float, r_c_deklaag: float, d70_m: float,
        gamma_korrel: float, v: float, theta: float, eta: float, g: float,
        gamma_water: float,
        ) -> Tuple[float, ...]:
    """Grenstoestandsfuncties volgens het WBI-model met grondwaterstroming
    conform model 4a.
    """
    L_voorland = pc_piping.calc_lengte_voorland(L_intrede=L_intrede,
                                                L_but=L_but)
    lambda_voorland = pc_piping.calc_lambda_voorland(kD_wvp=kD_wvp,
                                                     c_voorland=c_voorland)
    W_voorland = pc_piping.calc_W_voorland(lambda_voorland=lambda_voorland,
                                           L_voorland=L_voorland)
    L_kwelweg = pc_piping.calc_L_kwelweg(L_but=L_but, W_voorland=W_voorland)
    h_exit = pc_piping.calc_h_exit(polderpeil=polderpeil, mv_exit=mv_exit)
    r_exit = pc_piping.calc_r_exit_model4a(
        kD_wvp=kD_wvp, D_wvp=D_wvp, c_voorland=c_voorland,
        c_achterland=c_achterland, L_but=L_but, L_bit=L_bit,
        L_achterland=L_achterland, L_voorland=L_voorland)
    d_deklaag = pc_piping.calc_d_deklaag(mv_exit=mv_exit, top_zand=top_zand)
    phi_exit = pc_piping.calc_phi_exit(polderpeil=polderpeil, r_exit=r_exit,
                                       buitenwaterstand=buitenwaterstand)
    dphi_c_u = pc_piping.calc_dphi_c_u(
        d_deklaag=d_deklaag, gamma_sat_deklaag=gamma_sat_deklaag,
        gamma_water=gamma_water)
    i_exit = pc_piping.calc_i_exit(phi_exit=phi_exit, h_exit=h_exit,
                                   d_deklaag=d_deklaag)
    dh_c = pc_piping.calc_dh_c(
        d70=d70, D_wvp=D_wvp, kD_wvp=kD_wvp, L_kwelweg=L_kwelweg,
        gamma_water=gamma_water, g=g, v=v, theta=theta,
        eta=eta, d70_m=d70_m, gamma_korrel=gamma_korrel)
    dh_red = pc_piping.calc_dh_red(
        buitenwaterstand=buitenwaterstand, h_exit=h_exit,
        r_c_deklaag=r_c_deklaag, d_deklaag=d_deklaag)

    z_u = modelfactor_u * dphi_c_u - (phi_exit - h_exit)
    z_h = (modelfactor_h * i_c_h) - i_exit
    z_p = (modelfactor_p * modelfactor_ff * modelfactor_3d
           * modelfactor_aniso * modelfactor_ml * dh_c) - dh_red
    z_combin = max(z_u, z_h, z_p)

    return (z_u, z_h, z_p, z_combin, h_exit, r_exit, phi_exit, d_deklaag,
            dphi_c_u, i_exit, L_voorland, lambda_voorland,
            W_voorland, L_kwelweg, dh_c, dh_red)


# noinspection PyPep8Naming
def limit_state_moria(  # TODO: Naam moria vervangen voor iets generieks?
        # Geometry parameters
        # TODO: Moet L_but L_buk worden? Of L_spreidingslengte?
        L_intrede: float, L_but: float,
        # Boundary condition parameters
        buitenwaterstand: float, buitenwaterstand_gemiddeld: float,
        polderpeil: float, mv_exit: float,
        # Subsoil property parameters
        lambda_voorland: float, phi_exit_gemiddeld: float, r_exit: float,
        top_zand: float, k_wvp: float, D_wvp: float,
        d70: float, gamma_sat_deklaag: float,
        # Model property parameters
        modelfactor_u: float, modelfactor_h: float, modelfactor_p: float,
        modelfactor_ff: float, modelfactor_3d: float,
        modelfactor_aniso: float, modelfactor_ml: float, i_c_h: float,
        r_c_deklaag: float, d70_m: float, gamma_korrel: float, v: float,
        theta: float, eta: float,
        # Constants
        g: float, gamma_water: float,
) -> Tuple[float, ...]:
    """ Grenstoestandsfuncties volgens het WBI-model met
    grondwaterstroming conform MORIA model.

    :param L_intrede:
    :param L_but:
    :param buitenwaterstand:
    :param buitenwaterstand_gemiddeld:
    :param polderpeil:
    :param mv_exit:
    :param lambda_voorland:
    :param phi_exit_gemiddeld:
    :param r_exit:
    :param top_zand:
    :param k_wvp:
    :param D_wvp:
    :param d70:
    :param gamma_sat_deklaag:
    :param modelfactor_u:
    :param modelfactor_h:
    :param modelfactor_p:
    :param modelfactor_ff: Model factor fijne fractie.
    :param modelfactor_3d: Model factor 3D effecten.
    :param modelfactor_aniso: Model factor anisotropie.
    :param modelfactor_ml: Model factor meerlaagsheid zandpakket.
    :param i_c_h:
    :param r_c_deklaag:
    :param d70_m:
    :param gamma_korrel:
    :param v:
    :param theta:
    :param eta:
    :param g:
    :param gamma_water:
    :return:
    """

    kD_wvp = k_wvp * D_wvp
    L_voorland = pc_piping.calc_lengte_voorland(
        L_intrede=L_intrede, L_but=L_but)
    W_voorland = pc_piping.calc_W_voorland(
        lambda_voorland=lambda_voorland, L_voorland=L_voorland)
    L_kwelweg = pc_piping.calc_L_kwelweg(L_but=L_but, W_voorland=W_voorland)
    h_exit = pc_piping.calc_h_exit(polderpeil=polderpeil, mv_exit=mv_exit)
    d_deklaag = pc_piping.calc_d_deklaag(mv_exit=mv_exit, top_zand=top_zand)
    phi_exit = (
        phi_exit_gemiddeld + r_exit * (buitenwaterstand
                                       - buitenwaterstand_gemiddeld)
        )
    dphi_c_u = pc_piping.calc_dphi_c_u(
        d_deklaag=d_deklaag,
        gamma_sat_deklaag=gamma_sat_deklaag,
        gamma_water=gamma_water)
    i_exit = pc_piping.calc_i_exit(
        phi_exit=phi_exit, h_exit=h_exit, d_deklaag=d_deklaag)
    dh_c = pc_piping.calc_dh_c(
        d70=d70, D_wvp=D_wvp, kD_wvp=kD_wvp, L_kwelweg=L_kwelweg,
        gamma_water=gamma_water, g=g, v=v, theta=theta,
        eta=eta, d70_m=d70_m, gamma_korrel=gamma_korrel)
    dh_red = pc_piping.calc_dh_red(
        buitenwaterstand=buitenwaterstand, h_exit=h_exit,
        r_c_deklaag=r_c_deklaag, d_deklaag=d_deklaag)
    z_u = modelfactor_u * dphi_c_u - (phi_exit - h_exit)
    z_h = (modelfactor_h * i_c_h) - i_exit
    z_p = (modelfactor_p * modelfactor_ff * modelfactor_3d * modelfactor_aniso
           * modelfactor_ml * dh_c) - dh_red
    z_combin = max(z_u, z_h, z_p)

    return (z_u, z_h, z_p, z_combin, h_exit, phi_exit, d_deklaag,
            dphi_c_u, i_exit, L_voorland, W_voorland,
            L_kwelweg, kD_wvp, dh_c, dh_red)



from __future__ import annotations
import os
import pandas as pd
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from pandas import merge
from geoprob_pipe.calculations.systems.mappers.initial_input_mapper import INITIAL_INPUT_MAPPER
from typing import TYPE_CHECKING

from geoprob_pipe.cmd_app.utils.misc import get_geohydrological_model

if TYPE_CHECKING:
    from geoprob_pipe import GeoProbPipe


def overview_alpha(geoprob_pipe: GeoProbPipe, export: bool = False):

    model_string = get_geohydrological_model(app_settings=geoprob_pipe.input_data.app_settings)
    initial_input_mapper = INITIAL_INPUT_MAPPER[model_string]['input']

    # Get data for graphing
    df = geoprob_pipe.results.df_alphas_influence_factors_and_physical_values(
        filter_deterministic=False, filter_derived=False)
    df = df[["uittredepunt_id", "ondergrondscenario_id", "vak_id", "variable", "distribution_type", "physical_value"]]

    gdf_uittredepunten = geoprob_pipe.input_data.uittredepunten.gdf
    df = merge(df, gdf_uittredepunten[["uittredepunt_id", "metrering"]], on="uittredepunt_id", how="left")

    # Determine scenario order per uittredepunt_id
    df["scenario_order"] = (
        df.groupby("uittredepunt_id")["ondergrondscenario_id"].transform(lambda x: pd.factorize(x)[0] + 1))
    scenario_orders = sorted(df["scenario_order"].unique())

    # List of all alphas that can be shown
    parameters: list[str] = list(df["variable"].unique())
    # Add distribution type
    dist_types = {}
    for param in parameters:
        dist_type = df.loc[
            df["variable"] == param, "distribution_type"].unique()[0]
        dist_types.update({param: dist_type})
    # Add units
    unit_lookup = {item["name"]: item["unit"] for item in initial_input_mapper}
    param_units = {}
    for param in parameters:
        try:
            param_units.update({param: str(unit_lookup[param]).strip("[]")})
        except KeyError:
            param_units.update({param: "?"})

    # Add parameter units missing in DUMMY_INPUT by hand.
    param_units.update({
        "L_kwelweg": "m",
        "L_voorland": "m",
        "W_voorland": "s/m",
        "buitenwaterstand_gemiddeld": "m+NAP",
        "d_deklaag": "m",
        "dh_c": "m",
        "dh_red": "m",
        "dphi_c_u": "m+NAP",
        "h_exit": "m+NAP",
        "i_exit": "-",
        "k_wvp": "m/dag",
        "lambda_voorland": "m",
        "phi_exit": "m+NAP",
        "phi_exit_gemiddeld": "m+NAP",
        "r_exit": "-",
        "z_combin": "-",
        "z_h": "-",
        "z_p": "-",
        "z_u": "-"
    })

    # Create subplots
    fig = make_subplots(
        rows=len(parameters),
        cols=1,
        shared_xaxes=False,
        subplot_titles=parameters
    )

    # Add a button for each ondergrondscenario
    buttons = []

    # Add scatter plot per scenario
    for i, scen_order in enumerate(scenario_orders):
        df_case = df[df["scenario_order"] == scen_order]

        for row_idx, param in enumerate(parameters, start=1):
            df_param = df_case[df_case["variable"] == param]
            fig.add_trace(
                go.Scatter(
                    x=df_param["metrering"],
                    y=df_param["physical_value"],
                    mode="markers",
                    marker=dict(color="black", symbol="x", size=5),
                    name=param,
                    visible=(i == 0)
                ),
                row=row_idx, col=1
            )
            fig.update_xaxes(showgrid=True, tickangle=90,
                             row=row_idx, col=1)
            fig.update_yaxes(showgrid=True,
                             title_text=f"{param} [{param_units[param]}]"
                             + f"<br>({dist_types[param]})",
                             row=row_idx, col=1)

        total_traces = len(scenario_orders) * len(parameters)

        # Determine which scenario is visible
        vis = [False] * total_traces
        vis[i*len(parameters):(i+1)*len(parameters)] = [True] * len(parameters)

        buttons.append(dict(
            label=f"Scenario {scen_order}",
            method="update",
            args=[
                {"visible": vis},
                {"title": f"Overview of parameters for Scenario {scen_order}"}
            ]
        ))

    # Layout and button
    fig.update_layout(
        height=300*len(parameters),
        showlegend=False,
        title=f"Overview of parameters for Scenario {scenario_orders[0]}",
        updatemenus=[dict(
            active=0,
            buttons=buttons,
            direction="down",
            showactive=True,
            x=1.05, y=1.01
        )],
    )

    # Export
    if export:
        export_dir = os.path.join(
            geoprob_pipe.visualizations.graphs.export_dir,
            "grafiek_physical_values"
        )
        os.makedirs(export_dir, exist_ok=True)

        fig.write_html(
            os.path.join(export_dir, "overview_alphas.html"),
            include_plotlyjs='cdn'
        )

        for scen_order in scenario_orders:
            df_case = df[df["scenario_order"] == scen_order]
            fig_case = make_subplots(
                rows=len(parameters), cols=1,
                shared_xaxes=False,
                subplot_titles=parameters
            )
            for row_idx, param in enumerate(parameters, start=1):
                df_param = df_case[df_case["variable"] == param]
                fig_case.add_trace(
                    go.Scatter(
                        x=df_param["metrering"],
                        y=df_param["physical_value"],
                        mode="markers",
                        marker=dict(color="black", symbol="x", size=5),
                        name=param
                    ),
                    row=row_idx, col=1
                )
                fig_case.update_xaxes(showgrid=True, tickangle=90,
                                      row=row_idx, col=1)
                fig_case.update_yaxes(showgrid=True,
                                      title_text=f"{param} [{param_units[param]}]"
                                      + f"<br>({dist_types[param]})",
                                      row=row_idx, col=1)

            fig_case.update_layout(
                height=300*len(parameters),
                showlegend=False,
                title=f"Overview of parameters for Scenario {scen_order}"
            )

    return fig
